# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""LOM analysis based on https://arxiv.org/pdf/2103.10344.pdf
"""
# pylint: disable=invalid-name

# monkey patch to temporarily mock h5py dependency required by scqubits, which conflicts with
# geopandas
import sys

from scipy.fftpack import hilbert


# pylint: disable=wrong-import-position
# pylint: disable=too-few-public-methods
class DummyH5py:

    @property
    def Group(self):
        pass


sys.modules['h5py'] = DummyH5py

from collections import defaultdict, namedtuple
from typing import Any, List, Dict, Tuple, DefaultDict, Sequence, Mapping, Union, Callable
import argparse

import numpy as np
import pandas as pd
import scqubits as scq
from sympy import Matrix
from scipy import optimize, integrate

from pyEPR.calcs.convert import Convert
from pyEPR.calcs.constants import e_el as ele, hbar

from qiskit_metal.toolbox_python.utility_functions import get_all_args
from qiskit_metal.analyses.em.cpw_calculations import guided_wavelength
from qiskit_metal.analyses.hamiltonian.states_energies import extract_energies
from .constants import (MHzRad, GHzRad, NANO, FEMTO, ONE_OVER_FEMTO)

from qiskit_metal import logger

BasisTransform = namedtuple(
    'BasisTransform',
    ['orig_node_basis', 'node_jj_basis', 'num_negative_nodes'])

Operator = namedtuple('Operator', ['op', 'add_hc'])


class CouplingType:
    CAPACITIVE = 'CAPACITIVE'
    INDUCTIVE = 'INDUCTIVE'


def analyze_loaded_tl(fr, vp, Z0, cap_loading: Dict[str, float], shorted=False):
    """ Calculate charge operator zero point fluctuation, Q_zpf at the point of the loading
    capacitor.
    https://arxiv.org/pdf/2103.10344.pdf
    equation 20

    Args:
        fr (float): TL resonator frequency in MHz
        vp (float): phase velocity in m/s
        Z0 (float): characteristic impedance of the TL in ohm
        cap_loading (dict): a dictionary of the loading capacitors; the keys
            are the names of the nodes; the values the capacitances in fF
        shorted (boolean): default false; true if the other end of the TL is shorted false otherwise

    Returns:
        [type]: [description]
    """
    # An arbitrarily very large positive number in place of positive infinity.
    # Almost exclusively function as an infinity flag instead of actually
    # particupating in algebraic calculations, except for the boundary condition
    # of a shorted end. In that case, in the calculation of the capacitive energy
    # participation, the shorted end term is proportional to ,
    # (C_shorted * cos(pi/2)) ^ 2, where C_shorted is infinity. Here, one cannot
    # use np.infty for C_shorted because np.infty * np.cos(np.pi/2) = np.infty in python
    # instead of zero as one would expect.
    #
    # Ultimately, the energy particpation of the shorted end doesn't matter as
    # long as it's close to zero and not infinity since it's energy participation
    # of the other (loaded with finite capacitance) end that matters.
    _POS_INFTY = 1e30
    # Convert to SI
    wr = fr * MHzRad
    if cap_loading == {}:
        raise ValueError('At least one loading capacitor needs to be defined. ')
    elif len(cap_loading) == 1:
        cap_loading['_cl'] = 0 if not shorted else _POS_INFTY

    w_loading = {}
    for node, val in cap_loading.items():
        cap_loading[node] = val * FEMTO
        if val == _POS_INFTY:
            w_loading[node] = 0
        else:
            w_loading[node] = 1 / (Z0 *
                                   cap_loading[node]) if val else _POS_INFTY

    # assign z = 0 to one of the nodes and z = L to the other; switching
    # the assignment would not change the result
    z = {}
    nodes = set(cap_loading.keys())
    for node, val in cap_loading.items():
        z['0'] = node
        if val != 0 and val < _POS_INFTY:
            break
    z['L'] = list(nodes - {z['0']})[0]

    c = 1 / (Z0 * vp)  # cap/unit length

    phi = np.arctan(wr / w_loading[z['0']])
    k = wr / vp
    utl = lambda z: np.cos(k * z + phi)
    utl2 = lambda z: utl(z)**2
    m = 1  # mode number

    w_loading_vals = list(w_loading.values())

    def _arctan_arg(w_L, w_R):
        """Need separate handling for limiting cases
        from different boundary conditions
        """
        # left boundary of the TL is open, i.e., left loading capacitor is 0
        if w_L == _POS_INFTY:
            return -wr / w_R
        # right boundary of the TL is open, i.e., right loading capacitor is 0
        elif w_R == _POS_INFTY:
            return -wr / w_L
        else:
            return wr * (w_L + w_R) / (wr**2 - w_L * w_R)

    root_eq = lambda L: np.arctan(
        _arctan_arg(w_loading_vals[0], w_loading_vals[1])
    ) + m * np.pi - k * L  # = 0
    sol = optimize.root(root_eq, [0.007], jac=False,
                        method='hybr')  # in meters SI
    Ltl = sol.x[0]

    E_cap = (0.5 * c * integrate.quad(utl2, 0, Ltl)[0] +
             0.5 * cap_loading[z['0']] * utl(0)**2 +
             0.5 * cap_loading[z['L']] * utl(Ltl)**2)

    pCL = {}
    Q_zpf = {}
    Phi_zpf = {}

    for node, val in cap_loading.items():
        if node == z['0']:
            pCL[node] = 0.5 * val * utl(0)**2 / E_cap
        else:
            pCL[node] = 0.5 * val * utl(Ltl)**2 / E_cap
        Q_zpf[node] = np.sqrt(hbar * wr / 2 * pCL[node] * val)

        # using the uncertainty relationship that Q_zpf * Phi_zpf = hbar / 2
        Phi_zpf[node] = 0.5 * hbar / Q_zpf[node]

    return Q_zpf, Phi_zpf, phi, Ltl


def _product_no_duplicate(*args) -> List[List]:
    """
    product_no_duplicate('ABCDx', 'xy') --> Ax Ay Bx By Cx Cy Dx Dy xy
    ** Note that xx is not included **

    Similar to itertools.product except for results with duplicates are
    excluded

    """

    pools = [tuple(pool) for pool in args]
    result = [[]]

    for pool in pools:
        result = [x + [y] for x in result for y in pool if y not in x]

    return result


def _make_cmat_df(cmat, nodes):
    """
    generate a pandas dataframe from a capacitance matrix and list of node names
    """
    df = pd.DataFrame(cmat)
    df.columns = nodes
    df.index = nodes
    return df


def _df_cmat_to_adj_list(df_cmat: pd.DataFrame) -> DefaultDict(list):
    """
    generate an adjacency list from a capacitance matrix in a dataframe
    """
    nodes = df_cmat.columns.values
    vals = df_cmat.values
    graph = defaultdict(list)
    for ii, node in enumerate(nodes):
        for jj in range(ii, len(nodes)):
            graph[node].append((nodes[jj], vals[ii, jj]))
    return graph


def _cj_dict_to_adj_list(cj_dict: Dict[Tuple, float]) -> DefaultDict(list):
    graph = defaultdict(list)
    for (n1, n2), c in cj_dict.items():
        graph[n1].append((n2, -c))
        graph[n1].append((n1, c))
        graph[n2].append((n2, c))
    return graph


def _remove_row_col_at_idx(mat, idx):
    """
    remove row and column in a matrix corresponding an index
    """
    mat_ = np.delete(mat, idx, 0)
    mat_ = np.delete(mat_, idx, 1)
    return mat_


def _extract_matrix_from_transform(transform: BasisTransform,
                                   junctions: Mapping[Tuple[str, str], str],
                                   index: pd.Index) -> np.ndarray:
    """
    calculate the S_n^{-1} matrix that corresponds to a given transform
    https://arxiv.org/pdf/2103.10344.pdf
    """
    junctions = dict(zip(junctions.values(), junctions.keys()))

    dim = len(transform.orig_node_basis)
    m = np.zeros((dim, dim))
    for ii, (n_old, n_new) in enumerate(
            zip(transform.orig_node_basis, transform.node_jj_basis)):
        if n_old == n_new:
            m[ii, ii] = 1
        else:
            n1, n2 = junctions[n_new]
            c1, c2 = index.get_indexer([n1, n2])
            m[ii, c1] = 1
            m[ii, c2] = -1
    return m


def _transform_to_junction_flux_basis(orig_nodes,
                                      junctions: Mapping[Tuple[str, str], str],
                                      choose_least_num_neg: bool = True):
    """
    For example:
    junctions = {('n1', 'n2'):'j1', ('n3', 'n4'):'j2'}
    junction flux is defined such that Phi_{j1} = Phi_{1} - Phi_{2}, for {'j1': ('n1', 'n2')}, where we refer to n1 as
    the "positive" node and n2 as the "negative" node

    Param:
        choose_least_num_neg: if set true, choose the transform with least number of -1s in the transform matrix.
            this means choosing a basis which has the LARGEST number of "positive" junction nodes, i.e.,
            the LEAST number of "negative" junction nodes; if set false, return all possible transforms
    """
    junctions = dict(zip(junctions.values(), junctions.keys()))

    j_list = list(junctions.keys())

    # each junction flux can potentially replace one of two original nodes in the new flux basis.
    # for N junctions, an array of N nodes can represent such mapping where the ith node in the array
    # is the original node that is replaced by the ith junction; And an array of these arrays can
    # represent all possible mappings

    all_combos = _product_no_duplicate(*tuple(junctions.values()))
    all_maps = [dict(zip(x, j_list)) for x in all_combos]

    _transforms = []
    for _map in all_maps:
        new_basis = []
        num_negative_nodes = 0
        for n in orig_nodes:
            if n not in _map:
                new_basis.append(n)
            else:
                new_basis.append(_map[n])
                if junctions[_map[n]][1] == n:  # "negative" node
                    num_negative_nodes += 1

        _transforms.append(
            BasisTransform(orig_nodes, new_basis, num_negative_nodes))

    if choose_least_num_neg:
        _transforms.sort(key=lambda t: t.num_negative_nodes)
        return _transforms[0]

    return _transforms


class LabeledNdarray(np.ndarray):
    """np.ndarray subclass that shows labels when the array is printed.

    The constructor only accepts a square matrix. Sliced LabeledNdarray will still be
    a LabeledNdarray but the labels are not preserved: if the resulting matrix is still
    a square matrix, the labels will just be [0, 1, ...<array_row_number>-1]; and if the
    the resulting matrix is not a square matrix, there will be no labels.

    """

    def __new__(cls, input_array: np.ndarray, labels=None):
        if input_array.shape[0] != input_array.shape[1]:
            raise ValueError(
                'LabeledNdarray constructor only supports square matrices.')
        obj = np.asarray(input_array).view(cls)
        obj.labels = labels if labels is not None else [
            str(x) for x in range(obj.shape[0])
        ]
        return obj

    def to_dataframe(self):
        """ output to pandas dataframe
        """
        if len(self.shape) == 1 or self.shape[0] != self.shape[1]:
            raise ValueError(
                'Only square matrices are labeled and can be exported to dataframe.'
            )
        return _make_cmat_df(self, self.labels)

    def __array_finalize__(self, obj):
        if obj is None:
            return
        if self.shape != obj.shape:
            self.labels = [str(x) for x in range(self.shape[0])]
        else:
            self.labels = getattr(obj, 'labels',
                                  [str(x) for x in range(self.shape[0])])

    def __repr__(self):
        if len(self.shape) == 1 or self.shape[0] != self.shape[1]:
            return f'{self.view(np.ndarray)}'
        return f'{_make_cmat_df(self, self.labels)}'

    def __str__(self):
        if len(self.shape) == 1 or self.shape[0] != self.shape[1]:
            return f'{self.view(np.ndarray)}'
        return f'{_make_cmat_df(self, self.labels)}'


#----------------------------------------------------------------------------------------------------


def _maybe_remove_grd_node_then_cache(f):
    """function decorator that takes the matrix output
    of a function, i.e., f(), removes the ground node from
    the matrix and caches the result
    """

    def wrapper(self):
        _attr = f'_{f.__name__}'
        if getattr(self, _attr) is None:
            m = f(self)
            if self.ignore_grd_node:
                m = _remove_row_col_at_idx(m, self.grd_idx)
            setattr(self, _attr, m)
        else:
            m = getattr(self, _attr)
        return m

    return wrapper


class CircuitGraph:
    """
    class implementing the lumped model circuit analysis.
    Notations closely follow that of https://arxiv.org/pdf/2103.10344.pdf

    User shouldn't have to instantiate this class themself as it is
    created automatically by the CompositeSystem class

    The object corresponds to the entire composite system, incorporating inputs
    of all the cells provided
    """

    units = {'capacitance': 'fF', 'inductance': 'nH'}
    _IGNORE_GRD_NODE = True

    def __init__(
        self,
        nodes: Sequence,
        grd_node: str,
        cmats: List[pd.DataFrame],
        ind_lists: List[Dict[Tuple, float]],
        junctions: Mapping[Tuple[str, str], str],
        cj_dicts: List[Dict[Tuple, float]] = None,
        nodes_force_keep: Sequence = None,
        s_remove_provided: Union[np.ndarray, bool] = False,
        s_keep_provided: Union[np.ndarray, bool] = False,
    ):
        """Initialize the class with parameters specifying the circuit

        Args:
            nodes (Sequence): a list of the nodes
            grd_node (str): name of the ground node
            cmats (List[pd.DataFrame]): list of Maxwell capacitance matrices in pandas
                dataframes, one for each cell provided
            ind_lists (List[Dict[Tuple, float]]): list of dicts where each dict specifies
                the inductances in each cell. For the dict, the keys are tuples of specifying
                the two nodes between which the inductors lie.
                For example, [{('n1', 'n2'): 10}, {('n5', 'n7'): 13}] specifies that there
                is an inductor of 10 nH between node 'n1' and 'n2' for one cell and an
                inductor of 13 nH between node 'n5' and 'n7' for another cell
            junctions (Mapping[Tuple[str, str], str]): dict mapping original circuit
                nodes to custom-named junction nodes. This is the parameter informing
                the LOM analysis between which nodes the Josephson junctions are located.
                For example, {('n1', 'n2'): 'j1'} specifies that there is a junction
                between nodes 'n1' and 'n2' and named as 'j1'
            cj_dicts (List[Dict[Tuple, float]], optional): if provided, specifies
                the junction capacitances in each cell. Structure is the same as ind_lists.
                For the dict, the keys are tuples of specifying the two nodes between which
                the junctions lie. For example, [{('n1', 'n2'): 2}, None] specifies that there
                is an junction capacitance of 2 fF between node 'n1' and 'n2' for the first
                provided cell and None for the second provided cell. Defaults to None.
            nodes_force_keep (Sequence, optional): a list of nodes that will not be eliminated
                during L and C matrices reduction. Use this parameter to specify non-dynamic
                nodes (nodes that are connected capacitors only or inductors only) that should
                be preserved. If not specified (i.e., None), all non-dynamic nodes are eliminated
            s_remove_provided (np.ndarray): precalculated s_remove; see documentation on S_Remove. This is
                useful when a circuit's topology doesn't change hence s_remove doesn't change and can be
                reused
            s_keep_provided (np.ndarray): precalculated s_keep; see documentation on S_Keep. This is
                useful when a circuit's topology doesn't change hence s_keep doesn't change and can be
                reused
        """

        self.nodes = list(nodes)
        self.idx = pd.Index(self.nodes)
        self._grd_node = grd_node
        self.ignore_grd_node = CircuitGraph._IGNORE_GRD_NODE
        self.grd_idx = self.idx.get_indexer([grd_node])[0]
        self.nodes_force_keep = nodes_force_keep

        self._c_graphs = [_df_cmat_to_adj_list(m) for m in cmats]
        if cj_dicts is not None:
            for cj_dict in cj_dicts:
                if cj_dict is not None:
                    self._c_graphs.append(_cj_dict_to_adj_list(cj_dict))

        self._ind_lists = ind_lists
        self._junctions = junctions

        self._C_n = None
        self._L_n_inv = None

        # TODO: Handle cached properties dependencies more rigorously, i.e., set them to None, when
        # their dependencies change
        self._C_inv_k = None
        self._C_k = None
        self._L_inv_k = None

        self._S_n = None
        self._node_jj_basis = None

        self._s_remove_provided = s_remove_provided
        self._s_keep_provided = s_keep_provided

    def __str__(self):
        str_out = 'node_jj_basis:\n'
        str_out += '-------------\n'
        str_out += '\n'
        str_out += f'{self.node_jj_basis}\n'
        str_out += '\n'
        str_out += 'nodes_keep:\n'
        str_out += '-------------\n'
        str_out += '\n'
        str_out += f'{self.get_nodes_keep()}\n'
        str_out += '\n'
        str_out += '\n'
        str_out += 'L_inv_k (reduced inverse inductance matrix):\n'
        str_out += '-------------\n'
        str_out += '\n'
        str_out += f'{self.L_inv_k}\n'
        str_out += '\n'
        str_out += 'C_k (reduced capacitance matrix):\n'
        str_out += '-------------\n'
        str_out += '\n'
        str_out += f'{self.C_k}\n'
        str_out += '\n'

        return str_out

    def _adj_list_to_mat(self, adj_list):
        """ convert adjacency list representation of capacitance graph to
        a matrix representation
        """
        idx = self.idx
        dim = len(idx)
        mat = np.zeros((dim, dim))
        for n1 in adj_list:
            for n2, w in adj_list[n1]:
                r = idx.get_indexer([n1])[0]
                c = idx.get_indexer([n2])[0]
                mat[r, c] += w
                if r != c:
                    mat[c, r] += w
        return mat

    def _inductance_list_to_Linv_mat(self, ind_dict):
        """ convert inductance list to inductance inverse matrix
        """
        idx = self.idx
        dim = len(idx)
        mat = np.zeros((dim, dim))
        for (n1, n2), l in ind_dict.items():
            if n1 == n2:
                raise ValueError(
                    'inductance needs to be specified between two different nodes'
                )
            idx1, idx2 = idx.get_indexer([n1, n2])
            mat[idx1, idx2] = mat[idx2, idx1] = -1 / l
        for ii in range(len(mat)):
            mat[ii, ii] = -mat[ii].sum()
        return mat

    @property
    @_maybe_remove_grd_node_then_cache
    def C_n(self):
        """
        https://arxiv.org/pdf/2103.10344.pdf
        equation (5)
        """
        dim = len(self.idx)
        C_n = np.zeros((dim, dim))
        for c_g in self._c_graphs:
            C_n += self._adj_list_to_mat(c_g)
        return LabeledNdarray(C_n, self.orig_node_basis)

    @property
    @_maybe_remove_grd_node_then_cache
    def L_n_inv(self):
        """
        https://arxiv.org/pdf/2103.10344.pdf
        equation (5)
        """
        dim = len(self.idx)
        L_n_inv = np.zeros((dim, dim))
        for l_dict in self._ind_lists:
            L_n_inv += self._inductance_list_to_Linv_mat(l_dict)
        return LabeledNdarray(L_n_inv, self.orig_node_basis)

    @property
    @_maybe_remove_grd_node_then_cache
    def S_n(self):
        """ Linear transformation from original node basis to
        node-junction basis where all non-linear dipole fluxes are
        explicitly in the basis. The flux of the j-th dipole
        $\Phi_{j} = \Phi_{n2} - \Phi_{n1}$, where $\Phi_{n2}$ and
        $\Phi_{n1}$ are the node-to-datum fluxes of its two nodes
        .
        $\Phi = S_{n}^{-1}\Phi_n$
        """
        t = _transform_to_junction_flux_basis(self.nodes,
                                              self._junctions,
                                              choose_least_num_neg=True)
        S_n_inv = _extract_matrix_from_transform(t, self._junctions, self.idx)
        S_n = np.linalg.inv(S_n_inv)
        self._node_jj_basis = t.node_jj_basis
        return S_n

    @property
    def C(self):
        """Transformed capacitance matrix of the composite system
        """
        return LabeledNdarray(
            self.S_n.T.dot(self.C_n).dot(self.S_n), self.node_jj_basis)

    @property
    def L_inv(self):
        """Transformed inductance inverse matrix of the composite system
        """
        return LabeledNdarray(
            self.S_n.T.dot(self.L_n_inv).dot(self.S_n), self.node_jj_basis)

    @property
    def orig_node_basis(self):
        """ Original node basis consisting of node-to-datum fluxes,
        excluding the ground node
        """
        _orig_basis = self.nodes
        if self.ignore_grd_node:
            _orig_basis = [
                _node for _node in _orig_basis if _node != self._grd_node
            ]
        return _orig_basis

    @property
    def node_jj_basis(self):
        """ Node-junction basis which is linear transformation
        from original node basis to where all non-linear dipole fluxes are
        explicitly in the basis. The flux of the j-th dipole
        $\Phi_{j} = \Phi_{n2} - \Phi_{n1}$, where $\Phi_{n2}$ and
        $\Phi_{n1}$ are the node-to-datum fluxes of its two nodes
        .
        $\Phi = S_{n}^{-1}\Phi_n$
        """
        if self._node_jj_basis is None:
            _ = self.S_n
        _node_jj_basis = self._node_jj_basis
        if self.ignore_grd_node:
            _node_jj_basis = [
                _node for _node in _node_jj_basis if _node != self._grd_node
            ]
        return _node_jj_basis

    @property
    def S_remove(self):
        """ Eliminate coupler constraints due to the singularity of
        the inverse inductance matrix. Here we eliminate fluxes
        in the kernel space of the transformed inverse inductance
        matrix. These are nodes that are only touched by capacitors and are
        considered non-dynamic nodes.
        """
        if self._s_remove_provided is not False:
            return self._s_remove_provided
        nodes_force_keep = self.nodes_force_keep if self.nodes_force_keep else []
        force_keep_idx = pd.Index(
            self.node_jj_basis).get_indexer(nodes_force_keep)
        if np.where(force_keep_idx < 0)[0].size:
            bad_nodes = list(
                np.array(nodes_force_keep)[np.where(force_keep_idx < 0)])
            bad_nodes_str = ', '.join(bad_nodes)
            raise ValueError(
                f'nodes {bad_nodes_str} in input node_force_keep are not in the flux basis [self.node_jj_basis].'
            )

        L_inv = self.L_inv
        _s_r = []
        null_space = Matrix(L_inv).nullspace()
        for _v in null_space:
            v = _v / _v.norm()
            v = np.array(v).astype(np.float64)
            if np.count_nonzero(v) != 1:
                raise ValueError(
                    f'Nullspace column vector {v} has more than one non-zero element. \
                                 Only individual nodes in the current flux [see self.node_jj_basis] basis can be removed'
                )

            # if the node to be removed is in the list of nodes that are forced to be kept, don't remove
            if np.where(force_keep_idx == np.where(v)[0][0])[0].size:
                continue
            _s_r.append(v)

        return np.concatenate(_s_r, axis=1) if _s_r else None

    @property
    def S_keep(self):
        """ Complement of S_remove
        """
        # FIXME: currently assuming that S_keep can be solely constructed from
        # S_remove (which itself is constructed from the identity matrix) and the identity matrix
        if self._s_keep_provided is not False:
            return self._s_keep_provided
        S_remove = self.S_remove
        dim = self.L_inv.shape[0]
        eye = np.eye(dim)
        return eye[:, np.where(S_remove.T.sum(
            axis=0) == 0)[0]] if S_remove is not None else eye

    def get_nodes_keep(self):
        s_keep = self.S_keep
        dim = s_keep.shape[0]
        keep_idx = s_keep.T.dot(np.arange(dim)[:, np.newaxis])[:, 0].astype(int)
        return np.array(self.node_jj_basis)[keep_idx].tolist()

    def get_nodes_remove(self):
        s_remove = self.S_remove
        if s_remove is not None:
            dim = s_remove.shape[0]
            remove_idx = s_remove.T.dot(
                np.arange(dim)[:, np.newaxis])[:, 0].astype(int)
            return np.array(self.node_jj_basis)[remove_idx].tolist()
        else:
            return []

    @property
    def L_inv_k(self):
        """
        https://arxiv.org/pdf/2103.10344.pdf
        equation (7a)
        """
        if self._L_inv_k is not None:
            return self._L_inv_k
        s_keep = self.S_keep
        l_inv = self.L_inv
        self._L_inv_k = LabeledNdarray(
            s_keep.T.dot(l_inv).dot(s_keep), self.get_nodes_keep())
        return self._L_inv_k

    @property
    def C_k(self):
        """
        https://arxiv.org/pdf/2103.10344.pdf
        equation (7b)
        """
        if self._C_k is not None:
            return self._C_k
        s_k = self.S_keep
        s_r = self.S_remove
        c = self.C
        _inner = c.dot(s_r).dot(np.linalg.inv(s_r.T.dot(c.dot(s_r)))).dot(
            s_r.T.dot(c)) if s_r is not None else 0
        self._C_k = LabeledNdarray(
            s_k.T.dot(c - _inner).dot(s_k), self.get_nodes_keep())
        return self._C_k

    @property
    def C_inv_k(self):
        if self._C_inv_k is not None:
            return self._C_inv_k
        c_k = self.C_k
        if np.linalg.matrix_rank(c_k) < c_k.shape[0]:
            raise ValueError('C_k is rank deficient hence can\'t be inverted')
        self._C_inv_k = LabeledNdarray(np.linalg.inv(self.C_k),
                                       self.get_nodes_keep())
        return self._C_inv_k


#----------------------------------------------------------------------------------------------------


def _rename_nodes_in_df(node_rename: Dict, df: pd.DataFrame) -> pd.DataFrame:
    orig_nodes = df.columns.values.tolist()
    new_nodes = [
        n if n not in node_rename else node_rename[n] for n in orig_nodes
    ]
    return _make_cmat_df(df.values, new_nodes)


def _rename_nodes_in_dict(node_rename: Dict, dict_input: Dict[Tuple[str],
                                                              Any]) -> Dict:
    renamed = {}
    for nodes, val in dict_input.items():
        _renamed_key = tuple(
            [n if n not in node_rename else node_rename[n] for n in nodes])
        renamed[_renamed_key] = val
    return renamed


def _find_subsystem_mat_index(cg: CircuitGraph, subsystem, single_node=True):
    nodes_keep = cg.get_nodes_keep()
    reduced_mat_idx = pd.Index(nodes_keep)

    ss_nodes = subsystem.nodes
    idx = reduced_mat_idx
    node_idx = idx.get_indexer(ss_nodes)
    subsystem_idx = node_idx[node_idx >= 0]
    if subsystem_idx.size != 1 and single_node:
        raise ValueError(
            f'One and ONLY one node in the reduced node-junction basis should map to the {subsystem.sys_type} subsystem.'
        )
    elif subsystem_idx.size != node_idx.size:
        raise ValueError(
            'One or more subsystem nodes not found in the circuit\'s nodes.')
    return subsystem_idx


def _process_input_gs(gs):
    # TODO: to be implemented
    return gs


class QuantumSystemRegistry:
    _system_registry = {}

    @classmethod
    def add_to_registry(cls, name, builder):
        cls._system_registry[name] = builder

    @classmethod
    def registry(cls):
        return cls._system_registry


class Subsystem:
    """Class representing a subsystem that can typically be mapped to a quantum system with known
    solution, such as a transmon, a fluxonium or a quantum harmonic osccilator

    frequencies if entered in q_opts need to be in GHz
    """

    def __init__(self,
                 name: str,
                 sys_type: str,
                 nodes: List[str],
                 q_opts: dict = None):
        """Initialize the Subsystem object

        Args:
            name (str): name of the subsystem
            sys_type (str): type of the subsystem. Type can be one of the types implemented
                by one of the concrete QuantumBuilders (call QuantumSystemRegistry.registry()
                to see all the choices), which can also be custom-made by user
            nodes (List[str]): list of nodes the subsystem corresponds to. For example, for
                a transmon subsystem, this would just be a list of a single node which is the
                user-defined junction node in the Cell object
            q_opts (dict, optional): options relevant to the type of quantum subsystem. The
                Subsystem object builds the quantum subsystem using quantumfy() which calls
                make_quantum() of the corresponding QuantumBuilder. q_opts provides
                parameters needed by make_quantum(). In the case of qubit subsystems, make_quantum()
                takes advantage of the scQbuits package (https://scqubits.readthedocs.io).
                Consequently, the default options correspond to the default options of the qubit
                classes in scQubits. Defaults to None.
        """
        self.name = name
        self.sys_type = sys_type
        self.nodes = nodes
        self.q_opts = q_opts
        self._quantum_system = None
        self._h_params = defaultdict(dict)
        self.system_params = {}

    @property
    def h_params(self):
        """ Hamiltonian parameters of the subsystem
        """
        if self._h_params == {}:
            raise ValueError(
                f'Subsystem {self.name}\'s Hamiltonian parameters have not been calculated.'
            )
        return self._h_params

    @property
    def quantum_system(self):
        """ The quantum subsystem that the subsystem corresponds to. It
        is built by calling an associated builder passed in quantumfy()
        using the Visitor pattern
        """
        if self._quantum_system is None:
            raise ValueError(
                f'Subsystem {self.name}\'s quantum system has not been created.'
            )
        return self._quantum_system

    def quantumfy(self, quantum_builder) -> None:
        """ Building the corresponding quantum system with an
        associated builder
        """
        quantum_builder.make_quantum(self)


class _QuantumBuilderMeta(type):

    def __init__(cls, name, bases, attrs):
        # register the class in the QuantumSystemType registry
        # if it is not the base class, i.e. excluding QuantumBuilder
        if bases:
            sys_type = getattr(cls, 'system_type', None)
            if sys_type is None:
                raise AttributeError(
                    'You need to define a type string for your QuantumBuilder class by assigning a "system_type" class attribute.'
                    +
                    ' For example, the "system_type" for the TransmonBuilder is "TRANSMON", FluxoniumBuilder "FLUXONIUM".'
                )
            QuantumSystemRegistry.add_to_registry(sys_type, cls)

            logger.info(
                '%s with system_type %s registered to QuantumSystemRegistry',
                cls.__name__, sys_type)

        super().__init__(name, bases, attrs)


class QuantumBuilder(metaclass=_QuantumBuilderMeta):

    def __init__(self, cg: CircuitGraph):
        self.cg = cg
        self.nodes_keep = self.cg.get_nodes_keep()
        self.reduced_mat_idx = pd.Index(self.nodes_keep)

    def make_quantum(self, subsystem: Subsystem):
        raise NotImplementedError


class QuantumBuilderOptions(argparse.Namespace):

    def set_from_input(self, input_opts: Dict):
        if input_opts is not None and len(input_opts):
            for k, v in input_opts.items():
                setattr(self, k, v)

    def view_as(self, group: Union[str, type]):
        if isinstance(group, str):
            return getattr(self, group, None)
        elif isinstance(group, type):
            opts = {}
            args = get_all_args(group.__init__)
            for arg in args:
                opts[arg] = getattr(self, arg, None)
            return opts


def set_builder_options(func):

    def wrapper(self, subsystem):
        dflt_opts = getattr(self, 'default_opts', {})
        build_options = QuantumBuilderOptions(**dflt_opts)
        build_options.set_from_input(subsystem.q_opts)
        setattr(self, 'builder_options', build_options)
        func(self, subsystem)

    return wrapper


class TransmonBuilder(QuantumBuilder):
    """
    Concrete builder class for a type of quantum subsystem
    Each subsystem of a composite system calls make_quantum() method of this class
    to map LOM analysis results, the reduced L and C matrix elements to hamiltonian
    parameters, such as Ej, Ec and etc

    To create your own custom quantum builder, follow the patterns of the existing
    ones, such this one, or the other ones implemented in this library

        system_type (str): name of the quantum subsystem
        default_opts (dict): default options relevant to the type of quantum
            subsystem. Provide any default values for the paremeters that
            make_quantum() needs. In the case of qubit subsystems, make_quantum()
            takes advantage of the scQbuits package (https://scqubits.readthedocs.io).
            Consequently, the default options correspond to the default options of
            the qubit classes in scQubits
    """

    system_type = 'TRANSMON'
    default_opts = {'ng': 0.001, 'ncut': 22, 'truncated_dim': 10}

    # pylint: disable=protected-access
    # pylint: disable=no-member
    @set_builder_options
    def make_quantum(self, subsystem: Subsystem):
        """ concrete building function for the builder to build the quantum
        system based on default options and input options set for the
        Subsystem object
        """
        cg = self.cg
        l_inv_k = cg.L_inv_k
        c_inv_k = cg.C_inv_k

        subsystem_idx = _find_subsystem_mat_index(cg,
                                                  subsystem,
                                                  single_node=True)
        ss_idx = subsystem_idx[0]
        subsystem.system_params['subsystem_idx'] = subsystem_idx

        # EJ and EC are in MHz
        EJ = Convert.Ej_from_Lj(1 / l_inv_k[ss_idx, ss_idx])
        EC = Convert.Ec_from_Cs(1 / c_inv_k[ss_idx, ss_idx])
        Q_zpf = 2 * ele
        builder_options = self.builder_options
        builder_options.EJ = EJ
        builder_options.EC = EC
        transmon = scq.Transmon(**builder_options.view_as(scq.Transmon))

        subsystem._quantum_system = transmon
        node = subsystem.nodes[0]
        subsystem._h_params[node] = dict(EJ=EJ, EC=EC, Q_zpf=Q_zpf)
        subsystem._h_params[node]['default_charge_op'] = Operator(
            transmon.n_operator(), False)


class FluxoniumBuilder(QuantumBuilder):
    """
    Concrete builder class for a type of quantum subsystem
    Each subsystem of a composite system calls make_quantum() method of this class
    to map LOM analysis results, the reduced L and C matrix elements to hamiltonian
    parameters, such as Ej, Ec and etc

    To create your own custom quantum builder, follow the patterns of the existing
    ones, such this one, or the other ones implemented in this library

        system_type (str): name of the quantum subsystem
        default_opts (dict): default options relevant to the type of quantum
            subsystem. Provide any default values for the paremeters that
            make_quantum() needs. In the case of qubit subsystems, make_quantum()
            takes advantage of the scQbuits package (https://scqubits.readthedocs.io).
            Consequently, the default options correspond to the default options of
            the qubit classes in scQubits
    """

    system_type = 'FLUXONIUM'
    default_opts = {'cutoff': 110, 'truncated_dim': 10}

    # pylint: disable=protected-access
    # pylint: disable=no-member
    @set_builder_options
    def make_quantum(self, subsystem: Subsystem):
        """ concrete building function for the builder to build the quantum
        system based on default options and input options set for the
        Subsystem object
        """
        cg = self.cg
        c_inv_k = cg.C_inv_k

        subsystem_idx = _find_subsystem_mat_index(cg,
                                                  subsystem,
                                                  single_node=True)
        ss_idx = subsystem_idx[0]
        subsystem.system_params['subsystem_idx'] = subsystem_idx

        # EJ and EC are in MHz
        EC = Convert.Ec_from_Cs(1 / c_inv_k[ss_idx, ss_idx])
        Q_zpf = 2 * ele
        builder_options = self.builder_options
        builder_options.EC = EC
        fluxonium = scq.Fluxonium(**builder_options.view_as(scq.Fluxonium))

        subsystem._quantum_system = fluxonium

        node = subsystem.nodes[0]
        subsystem._h_params[node] = dict(EJ=builder_options.EJ,
                                         EC=builder_options.EC,
                                         EL=builder_options.EL,
                                         flux=builder_options.flux,
                                         Q_zpf=Q_zpf)
        subsystem._h_params[node]['default_charge_op'] = Operator(
            fluxonium.n_operator(), False)


class TLResonatorBuilder(QuantumBuilder):
    """
    Concrete builder class for a type of quantum subsystem
    Each subsystem of a composite system calls make_quantum() method of this class
    to map LOM analysis results, the reduced L and C matrix elements to hamiltonian
    parameters, such as Ej, Ec and etc

    To create your own custom quantum builder, follow the patterns of the existing
    ones, such this one, or the other ones implemented in this library

        system_type (str): name of the quantum subsystem
        default_opts (dict): default options relevant to the type of quantum
            subsystem. Provide any default values for the paremeters that
            make_quantum() needs. In the case of qubit subsystems, make_quantum()
            takes advantage of the scQbuits package (https://scqubits.readthedocs.io).
            Consequently, the default options correspond to the default options of
            the qubit classes in scQubits
    """

    system_type = 'TL_RESONATOR'
    default_opts = {
        'Z0': 50,
        'vp': 'use_design',  # phase velocity
        'design': {
            'line_width': 10 * 1e-6,
            'line_gap': 6 * 1e-6,
            'substrate_thickness': 750 * 1e-6,
            'film_thickness': 200 * 1e-9
        },
        'truncated_dim': 3,
        'other_end_shorted': False
    }

    # pylint: disable=protected-access
    # pylint: disable=no-member
    @set_builder_options
    def make_quantum(self, subsystem: Subsystem):
        """ concrete building function for the builder to build the quantum
        system based on default options and input options set for the
        Subsystem object
        """
        cg = self.cg
        c_inv_k = cg.C_inv_k

        builder_options = self.builder_options
        f_res = builder_options.f_res
        Z0 = builder_options.Z0
        # convert to MHz
        f_res *= 1000
        builder_options.E_osc = f_res

        if builder_options.vp == 'use_design':
            lambdaG, _, _ = guided_wavelength(f_res * 10**6,
                                              **builder_options.design)
            vp = f_res * MHzRad * lambdaG / (2 * np.pi)
        else:
            vp = float(builder_options.vp)

        subsystem_idx = _find_subsystem_mat_index(cg,
                                                  subsystem,
                                                  single_node=False)

        subsystem.system_params['subsystem_idx'] = subsystem_idx

        if len(subsystem_idx) > 2:
            raise ValueError(
                'Transmission line resonator currently supports at most two nodes.'
            )
        elif len(subsystem_idx) == 2 and builder_options.other_end_shorted:
            raise ValueError(
                'Transmission line resonator should have only one node \
                specified if one of its boundaries is shorted to ground.')

        loading_capacitor = {}
        for ii, node in enumerate(subsystem.nodes):
            loading_capacitor[node] = 1 / c_inv_k[subsystem_idx[ii],
                                                  subsystem_idx[ii]]

        resonator = scq.Oscillator(**builder_options.view_as(scq.Oscillator))
        subsystem._quantum_system = resonator

        Q_zpf, *_ = analyze_loaded_tl(f_res,
                                      vp,
                                      Z0,
                                      cap_loading=loading_capacitor,
                                      shorted=builder_options.other_end_shorted)

        for node in subsystem.nodes:
            subsystem._h_params[node]['Q_zpf'] = Q_zpf[node]
            subsystem._h_params[node]['default_charge_op'] = Operator(
                1j * (resonator.creation_operator() -
                      resonator.annihilation_operator()), False)


class LumpedResonatorBuilder(QuantumBuilder):
    """
    Concrete builder class for a type of quantum subsystem
    Each subsystem of a composite system calls make_quantum() method of this class
    to map LOM analysis results, the reduced L and C matrix elements to hamiltonian
    parameters, such as Ej, Ec and etc

    To create your own custom quantum builder, follow the patterns of the existing
    ones, such this one, or the other ones implemented in this library

        system_type (str): name of the quantum subsystem
        default_opts (dict): default options relevant to the type of quantum
            subsystem. Provide any default values for the paremeters that
            make_quantum() needs. In the case of qubit subsystems, make_quantum()
            takes advantage of the scQbuits package (https://scqubits.readthedocs.io).
            Consequently, the default options correspond to the default options of
            the qubit classes in scQubits
    """

    system_type = 'LUMPED_RESONATOR'
    default_opts = {'truncated_dim': 3}

    # pylint: disable=protected-access
    # pylint: disable=no-member
    @set_builder_options
    def make_quantum(self, subsystem: Subsystem):
        """ concrete building function for the builder to build the quantum
        system based on default options and input options set for the
        Subsystem object
        """
        cg = self.cg
        l_inv_k = cg.L_inv_k
        c_inv_k = cg.C_inv_k

        subsystem_idx = _find_subsystem_mat_index(cg,
                                                  subsystem,
                                                  single_node=True)
        ss_idx = subsystem_idx[0]
        subsystem.system_params['subsystem_idx'] = subsystem_idx

        builder_options = self.builder_options

        # TODO: should we be more explicit about units?
        # assume l_inv_k is in 1/nH and c_inv_k is in 1/fF
        l = 1 / l_inv_k[ss_idx, ss_idx] * NANO
        c = 1 / c_inv_k[ss_idx, ss_idx] * FEMTO

        # resonant angular frequency in radians/s
        w_res = 1 / np.sqrt(l * c)
        # convert to MHz
        f_res = w_res / MHzRad
        builder_options.E_osc = f_res

        Z = np.sqrt(l / c)
        Q_zpf = np.sqrt(hbar / (2 * Z))
        resonator = scq.Oscillator(**builder_options.view_as(scq.Oscillator))
        subsystem._quantum_system = resonator

        node = subsystem.nodes[0]
        subsystem._h_params[node]['Q_zpf'] = Q_zpf
        subsystem._h_params[node]['default_charge_op'] = Operator(
            1j *
            (resonator.creation_operator() - resonator.annihilation_operator()),
            False)


class Cell:
    """A physical subdivision of the device (i.e., the composite system), a sub-graph
    of capaciators, inductoros and other linear or non-linear circuit elements. The
    cell can be independently simulated to extract its electromagnetic parameters, which
    the the user provides as inputs to this object
    """

    def __init__(self, options: Dict):
        """Initialize the cell object

        Args:
            options (Dict): options can contain the following keys
                node_rename (dict): a dict mapping from original node names to
                    new node names, {old_name: new_name}
                cap_mat (pd.DataFrame): Maxwell capacitance of the cell in
                    pandas dataframe
                ind_dict (dict): the keys are tuples of specifying the two nodes
                    between which the inductors lie. For example,
                    {('n1', 'n2'): 10} specifies that there is an inductor of 10
                    nH between node 'n1' and 'n2' for one cell and an inductor of
                    13 nH between node 'n5' and 'n7' for another cell
                jj_dict (dict): dict mapping original circuit nodes to custom-named
                    junction nodes. This is the parameter informing the LOM analysis between
                    which nodes the Josephson junctions are located. For example,
                    {('n1', 'n2'): 'j1'} specifies that there is a junction between
                    nodes 'n1' and 'n2' and named as 'j1' cf_dict (dict):
                cj_dict (dict): if provided, specifies the junction capacitances in each
                    cell. Structure is the same as ind_dict. For the dict, the keys are
                    tuples of specifying the two nodes between which the junctions lie.
                    For example, {('n1', 'n2'): 2} specifies that there is an
                    junction capacitance of 2 fF between node 'n1' and 'n2' for the first
                    provided cell and None for the second provided cell.

        """
        self._node_rename = options.get('node_rename', {})
        self.cap_mat = _rename_nodes_in_df(self._node_rename,
                                           options['cap_mat'])
        self.ind_dict = None
        self.jj_dict = None
        self.cj_dict = None

        if 'ind_dict' in options:
            self.ind_dict = _rename_nodes_in_dict(self._node_rename,
                                                  options['ind_dict'])
        if 'ind_mat' in options:
            raise NotImplementedError
        if 'jj_dict' in options:
            self.jj_dict = _rename_nodes_in_dict(self._node_rename,
                                                 options['jj_dict'])

        if 'cj_dict' in options:
            self.cj_dict = _rename_nodes_in_dict(self._node_rename,
                                                 options['cj_dict'])

        self.nodes = self.cap_mat.columns.values.tolist()


class CompositeSystem:
    """Class representing the composite system which may consist of multiple subsystems and cells
    """

    def __init__(
        self,
        subsystems: List[Subsystem],
        cells: List[Cell],
        grd_node: str,
        nodes_force_keep: Sequence = None,
        s_remove_provided: Union[np.ndarray, bool] = False,
        s_keep_provided: Union[np.ndarray, bool] = False,
    ):
        """Initialize the CompositeSystem object

        Args:
            subsystems (List[Subsystem]): list of Subsystem objects
            cells (List[Cell]): list of Cell objects
            grd_node (str): name of the ground node
            nodes_force_keep (Sequence, optional): a list of nodes that will not be eliminated
                during L and C matrices reduction. Use this parameter to specify non-dynamic
                nodes (nodes that are connected capacitors only or inductors only) that should
                be preserved. If not specified (i.e., None), all non-dynamic nodes are eliminated
            s_remove_provided (np.ndarray): precalculated s_remove; see documentation on CircuitGraph.S_Remove
            s_keep_provided (np.ndarray): precalculated s_keep; see documentation on CircuitGraph.S_Keep
        """
        self._subsystems = subsystems
        self.num_subsystems = len(subsystems)
        self.names = [sub.name for sub in self._subsystems]
        self._cells = cells
        self.num_cells = len(cells)
        self.grd_node = grd_node
        self.nodes_force_keep = nodes_force_keep

        self._s_remove_provided = s_remove_provided
        self._s_keep_provided = s_keep_provided

        self._jj = {k: j for c in self._cells for k, j in c.jj_dict.items()}
        _jj_to_node_map = dict(zip(self._jj.values(), self._jj.keys()))

        _cell_nodes = np.unique([n for c in self._cells for n in c.nodes
                                ]).tolist()
        _sys_nodes = []
        for sub in self._subsystems:
            for n in sub.nodes:
                if n not in _jj_to_node_map:
                    _sys_nodes.append(n)
                else:
                    _sys_nodes.extend(
                        list(filter(lambda x: x != grd_node,
                                    _jj_to_node_map[n])))
        if len(set(_sys_nodes)) != len(_sys_nodes):
            raise ValueError('Subsystems should not have overlapping nodes.')
        _non_sys_nodes = [n for n in _cell_nodes if n not in _sys_nodes]
        self._nodes = _sys_nodes + _non_sys_nodes

        self._c_list = [c.cap_mat for c in self._cells]
        self._cj_dicts = [c.cj_dict for c in self._cells]
        self._l_list = [c.ind_dict for c in self._cells]

        self.quantum_subsystems = []

        self._cg = None

    def circuitGraph(self) -> CircuitGraph:
        """create a CircuitGraph object with circuit parameters of the composite system

        Returns:
            CircuitGraph: CircuitGraph object for LOM analysis
        """
        if self._cg is None:
            nodes = self._nodes
            grd_node = self.grd_node
            c_list = self._c_list
            cj_dicts = self._cj_dicts
            l_list = self._l_list
            jjs = self._jj
            self._cg = CircuitGraph(nodes,
                                    grd_node,
                                    c_list,
                                    l_list,
                                    jjs,
                                    cj_dicts,
                                    self.nodes_force_keep,
                                    s_keep_provided=self._s_keep_provided,
                                    s_remove_provided=self._s_remove_provided)

        return self._cg

    @property
    def subsystems(self):
        return self._subsystems

    def node_index(self, node: str):
        """Obtain the index of a given node in the composite system's reduced
        capacitance matrix

        Args:
            node (str): name of the node
        """
        cg = self.circuitGraph()
        nodes_keep = cg.get_nodes_keep()
        reduced_mat_idx = pd.Index(nodes_keep)

        node_idx = reduced_mat_idx.get_indexer([node])
        subsystem_idx = node_idx[node_idx >= 0]

        if subsystem_idx.size == 0:
            raise ValueError('Subsystem not found in the circuit\'s nodes.')
        return subsystem_idx[0]

    def create_hilbertspace(self) -> scq.HilbertSpace:
        """ create the composite hilbertspace including all the subsystems. Interaction
            NOT included

        Returns:
            scq.HilbertSpace: Hilbertspace object for the Hamiltonian
                of the composite system without interations added
        """
        cg = self.circuitGraph()

        quantum_systems = []
        for sub in self._subsystems:
            if sub.sys_type in QuantumSystemRegistry.registry():
                q_builder = QuantumSystemRegistry.registry()[sub.sys_type](cg)
            else:
                raise NotImplementedError

            sub.quantumfy(q_builder)
            quantum_systems.append(sub.quantum_system)

        self.quantum_subsystems = quantum_systems
        return scq.HilbertSpace(quantum_systems)

    def compute_gs(self, coupling_type: str = CouplingType.CAPACITIVE):
        """compute g matrices from reduced C inverse matrix C^{-1}_{k} and reduced L inverse matrix L^{-1}_{k}
        and zero point flucuations (Q_zpf, Phi_zpf)

        Note that the computed g's are in MHz

        Note: the resulting matrix is in the basis of nodes_keep, which may not be in the same order of
        subsystems
        """
        cg = self.circuitGraph()
        nodes = cg.get_nodes_keep()

        if coupling_type == CouplingType.CAPACITIVE:
            c_inv_k = cg.C_inv_k
            g = np.zeros_like(c_inv_k)
            for ii in range(self.num_subsystems):
                sub1 = self._subsystems[ii]
                sub1_nodes = sub1.nodes
                for jj in range(ii, self.num_subsystems):
                    sub2 = self._subsystems[jj]
                    sub2_nodes = sub2.nodes
                    for node1 in sub1_nodes:
                        for node2 in sub2_nodes:
                            if node1 == node2:
                                continue
                            idx1 = self.node_index(node1)
                            idx2 = self.node_index(node2)
                            c_i = c_inv_k[idx1, idx2]
                            if c_i == 0:
                                continue
                            Q_zpf_1 = sub1.h_params[node1]['Q_zpf']
                            Q_zpf_2 = sub2.h_params[node2]['Q_zpf']
                            g[idx1, idx2] = c_i * Q_zpf_1 * Q_zpf_2 * \
                                            ONE_OVER_FEMTO / hbar / MHzRad
                            g[idx2, idx1] = g[idx1, idx2]

            return LabeledNdarray(g, nodes)
        else:
            raise NotImplementedError

    def add_interaction(self,
                        gs: np.ndarray = None,
                        gscale: float = 1.) -> scq.HilbertSpace:
        """ add interaction terms to the composite hilbertspace

        Args:
            gs (np.ndarray): coupling strength matrix. If not provided,
                it will be calculated by compute_gs()
            gscale (float): coupling strength scale

        Returns:
            scq.HilbertSpace: Hilbertspace object for the Hamiltonian
                of the composite system with interations added
        """
        cg = self.circuitGraph()
        h = self.create_hilbertspace()

        if gs is None:
            gs = self.compute_gs(coupling_type=CouplingType.CAPACITIVE)
        else:
            gs = _process_input_gs(gs)

        c_inv_k = cg.C_inv_k
        l_inv_k = cg.L_inv_k

        if c_inv_k.shape != gs.shape:
            raise ValueError(
                'The dimension of g matrix doesn\'t match that of the reduced capacitance matrix.'
            )

        for ii in range(self.num_subsystems):
            sub1 = self._subsystems[ii]
            sub1_nodes = sub1.nodes
            for jj in range(ii, self.num_subsystems):
                sub2 = self._subsystems[jj]
                sub2_nodes = sub2.nodes
                for node1 in sub1_nodes:
                    for node2 in sub2_nodes:
                        if node1 == node2:
                            continue

                        idx1 = self.node_index(node1)
                        idx2 = self.node_index(node2)

                        g = gs[idx1, idx2]
                        if g == 0:
                            continue
                        q1 = sub1.quantum_system
                        q2 = sub2.quantum_system

                        if l_inv_k[idx1, idx2] != 0:
                            raise NotImplementedError(
                                'Interaction term for inductive coupling not yet implemented.'
                            )
                        if c_inv_k[idx1, idx2] != 0:
                            q1_op, add_hc1 = sub1.h_params[node1][
                                'default_charge_op']
                            q2_op, add_hc2 = sub2.h_params[node2][
                                'default_charge_op']

                            h.add_interaction(check_validity=False,
                                              g=g * gscale,
                                              op1=(q1_op, q1),
                                              op2=(q2_op, q2),
                                              add_hc=add_hc1 or add_hc2)
        return h

    def hamiltonian_results(self,
                            hilbertspace: scq.HilbertSpace,
                            evals_count=None,
                            print_info=True) -> pd.DataFrame:
        """Print and return results

        Args:
            hilbertspace (scq.HilbertSpace): Hilbertspace object for the Hamiltonian
                of the composite system
            evals_count (int, optional): Number of eigenenergy levels to keep
                after diagonalizing the Hamiltonian. Defaults to 10.
            print_info (bool, optional): If true, print results as well. Defaults to True.

        Returns:
            pd.DataFrame: dataframe containing the results
        """
        ham_res = {}

        names = self.names

        if evals_count is None:
            evals_count = hilbertspace.dimension

        hamiltonian_mat = hilbertspace.hamiltonian()
        evals, evecs = hamiltonian_mat.eigenstates(eigvals=evals_count)

        # evals, evecs = hilbertspace.eigensys(evals_count=evals_count)
        esys_array = np.empty(shape=(2,), dtype=object)
        esys_array[0] = evals
        esys_array[1] = evecs

        f01s, chi_mat = extract_energies(esys_array,
                                         mode_size=hamiltonian_mat.dims[0])
        f01s = f01s / 1000
        ham_res['fQ_in_Ghz'] = dict(zip(names, f01s))
        ham_res['chi_in_MHz'] = LabeledNdarray(chi_mat, names)

        if print_info:
            print('')
            print('system frequencies in GHz:')
            print('--------------------------')
            print(ham_res['fQ_in_Ghz'])
            print()
            print('Chi matrices in MHz')
            print('--------------------------')
            print(ham_res['chi_in_MHz'])
        return ham_res



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from typing import List, Tuple, Union

from numpy.linalg import norm

import numpy as np
from qiskit_metal import Dict
from qiskit_metal.toolbox_metal.parsing import is_true
from qiskit_metal.qlibrary.core import QRoute, QRoutePoint
from qiskit_metal.toolbox_metal import math_and_overrides as mao
from qiskit_metal.toolbox_metal.exceptions import QiskitMetalDesignError


class RouteMeander(QRoute):
    """Implements a simple CPW, with a single meander.  The base `CPW
    meandered` class.

    Inherits `QRoute` class

    .. meta::
        :description: Route Meander

    QRoute Default Options:
        * pin_inputs: Dict
            * start_pin: Dict -- Component and pin string pair. Define which pin to start from
                * component: '' -- Name of component to start from, which has a pin
                * pin: '' -- Name of pin used for pin_start
            * end_pin=Dict -- Component and pin string pair. Define which pin to start from
                * component: '' -- Name of component to end on, which has a pin
                * pin: '' -- Name of pin used for pin_end
        * fillet: '0'
        * lead: Dict
            * start_straight: '0mm' -- Lead-in, defined as the straight segment extension from start_pin.  Defaults to 0.1um.
            * end_straight: '0mm' -- Lead-out, defined as the straight segment extension from end_pin.  Defaults to 0.1um.
            * start_jogged_extension: '' -- Lead-in, jogged extension of lead-in. Described as list of tuples
            * end_jogged_extension: '' -- Lead-out, jogged extension of lead-out. Described as list of tuples
        * total_length: '7mm'
        * trace_width: 'cpw_width' -- Defines the width of the line.  Defaults to 'cpw_width'.

    Default Options:
        * meander: Dict
            * spacing: '200um' -- Minimum spacing between adjacent meander curves.  Defaults to 200um.
            * asymmetry='0um' -- offset between the center-line of the meander and the center-line that stretches from the tip of lead-in to the x (or y) coordinate of the tip of the lead-out.  Defaults to '0um'.
        * snap: 'true'
        * prevent_short_edges: 'true'
    """

    component_metadata = Dict(short_name='cpw')
    """Component metadata"""

    default_options = Dict(meander=Dict(spacing='200um', asymmetry='0um'),
                           snap='true',
                           prevent_short_edges='true')
    """Default options"""

    TOOLTIP = """Implements a simple CPW, with a single meander."""

    def make(self):
        """The make function implements the logic that creates the geometry
        (poly, path, etc.) from the qcomponent.options dictionary of
        parameters, and the adds them to the design, using
        qcomponent.add_qgeometry(...), adding in extra needed information, such
        as layer, subtract, etc."""
        # parsed options
        snap = is_true(self.p.snap)

        # Set the CPW pins and add the points/directions to the lead-in/out arrays
        self.set_pin("start")
        self.set_pin("end")

        # Align the lead-in/out to the input options set from the user
        meander_start_point = self.set_lead("start")
        meander_end_point = self.set_lead("end")

        # approximate length needed for the meander
        self._length_segment = self.p.total_length - (self.head.length +
                                                      self.tail.length)

        arc_pts = self.connect_meandered(meander_start_point, meander_end_point)

        self.intermediate_pts = arc_pts

        self.intermediate_pts = self.adjust_length(
            self.p.total_length - self.length, arc_pts, meander_start_point,
            meander_end_point)

        # Make points into elements
        self.make_elements(self.get_points())

    def connect_meandered(self, start_pt: QRoutePoint,
                          end_pt: QRoutePoint) -> np.ndarray:
        """Meanders using a fixed length and fixed spacing.

        Args:
            start_pt (QRoutePoint): QRoutePoint of the start
            end_pt (QRoutePoint): QRoutePoint of the end

        Returns:
            np.ndarray: Array of points

        Adjusts the width of the meander:
            * Includes the start but not the given end point
            * If it cannot meander just returns the initial start point
        """

        ################################################################
        # Setup

        # Parameters
        meander_opt = self.p.meander
        spacing = meander_opt.spacing  # Horizontal spacing between meanders
        asymmetry = meander_opt.asymmetry
        snap = is_true(self.p.snap)  # snap to xy grid
        prevent_short_edges = is_true(self.p.prevent_short_edges)

        # take care of anchors (do not have set directions)
        anchor_lead = 0
        if end_pt.direction is None:
            # end_direction originates strictly from endpoint + leadout (NOT intermediate stopping anchors)
            self.assign_direction_to_anchor(start_pt, end_pt)
            anchor_lead = spacing

        # Meander length
        length_meander = self._length_segment
        if self.p.snap:
            # handle y distance
            length_meander -= 0  # (end.position - endm.position)[1]

        # Coordinate system (example: x to the right => sideways up)
        forward, sideways = self.get_unit_vectors(start_pt, end_pt, snap)

        # Calculate lengths and meander number
        dist = end_pt.position - start_pt.position
        if snap:
            length_direct = abs(norm(mao.dot(
                dist, forward)))  # in the vertical direction
            length_sideways = abs(norm(mao.dot(
                dist, sideways)))  # in the orthogonal direction
        else:
            length_direct = norm(dist)
            length_sideways = 0

        # Breakup into sections
        meander_number = np.floor(length_direct / spacing)
        if meander_number < 1:
            self.logger.info(f'Zero meanders for {self.name}')
            return np.empty((0, 2), float)

        # The start and end points can have 4 directions each. Depending on the direction
        # there might be not enough space for all the meanders, thus here we adjust
        # meander_number w.r.t. what the start and end points "directionality" allows
        if mao.round(
                mao.dot(start_pt.direction, sideways) *
                mao.dot(end_pt.direction, sideways)) > 0 and (meander_number %
                                                              2) == 0:
            # even meander_number is no good if roots have same orientation (w.r.t sideway)
            meander_number -= 1
        elif mao.round(
                mao.dot(start_pt.direction, sideways) *
                mao.dot(end_pt.direction, sideways)) < 0 and (meander_number %
                                                              2) == 1:
            # odd meander_number is no good if roots have opposite orientation (w.r.t sideway)
            meander_number -= 1

        # should the first meander go sideways or counter sideways?
        start_meander_direction = mao.dot(start_pt.direction, sideways)
        end_meander_direction = mao.dot(end_pt.direction, sideways)
        if start_meander_direction > 0:  # sideway direction
            first_meander_sideways = True
        elif start_meander_direction < 0:  # opposite to sideway direction
            first_meander_sideways = False
        else:
            if end_meander_direction > 0:  # sideway direction
                first_meander_sideways = ((meander_number % 2) == 1)
            elif end_meander_direction < 0:  # opposite to sideway direction
                first_meander_sideways = ((meander_number % 2) == 0)
            else:
                # either direction is fine, so let's just pick one
                first_meander_sideways = True

        # length to distribute on the meanders (excess w.r.t a straight line between start and end)
        length_excess = (length_meander - length_direct - 2 * abs(asymmetry))
        # how much meander offset from center-line is needed to accommodate the length_excess (perpendicular length)
        length_perp = max(0, length_excess / (meander_number * 2.))

        # USES ROW Vectors
        # const vec. of unit normals
        middle_points = [forward] * int(meander_number + 1)
        # index so to multiply other column - creates a column vector
        scale_bys = spacing * np.arange(int(meander_number + 1))[:, None]
        # multiply each one in a linear chain fashion fwd
        middle_points = scale_bys * middle_points
        '''
        middle_points = array([
            [0. , 0. ],
            [0.2, 0. ],
            [0.4, 0. ],
            [0.6, 0. ],
            [0.8, 0. ],
            [1. , 0. ]])
        '''

        ################################################################
        # Calculation
        # including start and end points - there is no overlap in points
        # root_pts = np.concatenate([middle_points,
        #                            end.position[None, :]],  # convert to row vectors
        #                           axis=0)
        side_shift_vecs = np.array([sideways * length_perp] *
                                   len(middle_points))
        asymmetry_vecs = np.array([sideways * asymmetry] * len(middle_points))
        root_pts = middle_points + asymmetry_vecs
        top_pts = root_pts + side_shift_vecs
        bot_pts = root_pts - side_shift_vecs

        ################################################################
        # Combine points
        # Meanest part of the meander

        # Add 2 for the lead and end points in the cpw from
        # pts will have to store properly alternated top_pts and bot_pts
        # it will also store right-most root_pts (end)
        # 2 points from top_pts and bot_pts will be dropped for a complete meander
        pts = np.zeros((len(top_pts) + len(bot_pts) + 1 - 2, 2))
        # need to add the last root_pts in, because there could be a left-over non-meandered segment
        pts[-1, :] = root_pts[-1, :]
        idx_side1_meander, odd = self.get_index_for_side1_meander(len(root_pts))
        idx_side2_meander = 2 + idx_side1_meander[:None if odd else -2]
        if first_meander_sideways:
            pts[idx_side1_meander, :] = top_pts[:-1 if odd else None]
            pts[idx_side2_meander, :] = bot_pts[1:None if odd else -1]
        else:
            pts[idx_side1_meander, :] = bot_pts[:-1 if odd else None]
            pts[idx_side2_meander, :] = top_pts[1:None if odd else -1]

        pts += start_pt.position  # move to start position

        if snap:
            if ((mao.dot(start_pt.direction, end_pt.direction) < 0) and
                (mao.dot(forward, start_pt.direction) <= 0)):
                # pins are pointing opposite directions and diverging
                # the last root_pts need to be sideways aligned with the end.position point
                # and forward aligned with the previous meander point
                pts[-1, abs(forward[0])] = pts[-2, abs(forward[0])]
                pts[-1,
                    abs(forward[0]) - 1] = end_pt.position[abs(forward[0]) - 1]
            else:
                # the last root_pts need to be forward aligned with the end.position point
                pts[-1, abs(forward[0])] = end_pt.position[abs(forward[0])]
                # and if the last root_pts ends outside the CPW amplitude on the side where the last meander is
                # then the last meander needs to be locked on it as well
                if (self.issideways(pts[-1], pts[-3], pts[-2])
                        and self.issideways(pts[-2], root_pts[0]+start_pt.position, root_pts[-1]+start_pt.position))\
                        or (not self.issideways(pts[-1], pts[-3], pts[-2])
                            and not self.issideways(pts[-2], root_pts[0]+start_pt.position,
                                                    root_pts[-1]+start_pt.position)):
                    pts[-2, abs(forward[0])] = end_pt.position[abs(forward[0])]
                    pts[-3, abs(forward[0])] = end_pt.position[abs(forward[0])]
        if abs(asymmetry) > abs(length_perp):
            if not ((mao.dot(start_pt.direction, end_pt.direction) < 0) and
                    (mao.dot(forward, start_pt.direction) <= 0)):
                # pins are "not" pointing opposite directions and diverging
                if start_meander_direction * asymmetry < 0:  # sideway direction
                    pts[0, abs(forward[0])] = start_pt.position[abs(forward[0])]
                    pts[1, abs(forward[0])] = start_pt.position[abs(forward[0])]
                if end_meander_direction * asymmetry < 0:  # opposite sideway direction
                    pts[-2, abs(forward[0])] = end_pt.position[abs(forward[0])]
                    pts[-3, abs(forward[0])] = end_pt.position[abs(forward[0])]

        # Adjust the meander to eliminate the terminating jog (dogleg)
        if prevent_short_edges:
            x2fillet = 2 * self.p.fillet
            # adjust the tail first
            # the meander algorithm adds a final point in line with the tail, to cope with left-over
            # this extra point needs to be moved or not, depending on the tail tip direction
            if abs(mao.dot(end_pt.direction, sideways)) > 0:
                skippoint = 0
            else:
                skippoint = 1
            if 0 < abs(mao.round(end_pt.position[0] - pts[-1, 0])) < x2fillet:
                pts[-1 - skippoint,
                    0 - skippoint] = end_pt.position[0 - skippoint]
                pts[-2 - skippoint,
                    0 - skippoint] = end_pt.position[0 - skippoint]
            if 0 < abs(mao.round(end_pt.position[1] - pts[-1, 1])) < x2fillet:
                pts[-1 - skippoint,
                    1 - skippoint] = end_pt.position[1 - skippoint]
                pts[-2 - skippoint,
                    1 - skippoint] = end_pt.position[1 - skippoint]
            # repeat for the start. here we do not have the extra point
            if 0 < abs(mao.round(start_pt.position[0] - pts[0, 0])) < x2fillet:
                pts[0, 0] = start_pt.position[0]
                pts[1, 0] = start_pt.position[0]
            if 0 < abs(mao.round(start_pt.position[1] - pts[0, 1])) < x2fillet:
                pts[0, 1] = start_pt.position[1]
                pts[1, 1] = start_pt.position[1]

        return pts

    def adjust_length(self, delta_length, pts, start_pt: QRoutePoint,
                      end_pt: QRoutePoint) -> np.ndarray:
        """Edits meander points to redistribute the length slacks accrued with
        the various local adjustments It should be run after
        self.pts_intermediate is completely defined Inputs are however specific
        to the one meander segment Assumption is that pts is always a sequence
        of paired points, each corresponds to one meander 180deg curve The pts
        is typically an odd count since the last point is typically used to
        anchor the left-over length, therefore this code supports both odd and
        even cases, separately. For even it assumes all points are in paired.

        Args:
            delta_length (delta_length): slack/excess length to distribute on the pts
            pts (np.array): intermediate points of meander. pairs, except last point (2,2,...,2,1)
            start_pt (QRoutePoint): QRoutePoint of the start
            end_pt (QRoutePoint): QRoutePoint of the end

        Returns:
            np.ndarray: Array of points
        """
        # the adjustment length has to be computed in the main or in other method
        # considering entire route (Could include the corner fillet)

        if len(pts) <= 3:
            # not a meander
            return pts

        # is it an even or odd count of points?
        term_point = len(pts) % 2

        # recompute direction
        snap = is_true(self.p.snap)  # snap to xy grid
        forward, sideways = self.get_unit_vectors(start_pt, end_pt, snap)
        # recompute meander_sideways
        if mao.cross(pts[1] - pts[0], pts[2] - pts[1]) < 0:
            first_meander_sideways = True
        else:
            first_meander_sideways = False
        if mao.cross(pts[-2 - term_point] - pts[-1 - term_point],
                     pts[-3 - term_point] - pts[-2 - term_point]) < 0:
            last_meander_sideways = False
        else:
            last_meander_sideways = True

        # which points need to receive the shift?
        # 1. initialize the shift vector to 1 (1 = will receive shift)
        adjustment_vector = np.ones(len(pts))
        # 2. switch shift direction depending on sideways or not
        if first_meander_sideways:
            adjustment_vector[2::4] *= -1
            adjustment_vector[3::4] *= -1
        else:
            adjustment_vector[::4] *= -1
            adjustment_vector[1::4] *= -1

        # 3. suppress shift for points that can cause short edges
        # calculate thresholds for suppression of short edges (short edge = not long enough for set fillet)
        fillet_shift = sideways * self.p.fillet
        start_pt_adjusted_up = start_pt.position + fillet_shift
        start_pt_adjusted_down = start_pt.position - fillet_shift
        end_pt_adjusted_up = end_pt.position + fillet_shift
        end_pt_adjusted_down = end_pt.position - fillet_shift

        # if start_pt.position is below axes + shift - 2xfillet &  first_meander_sideways
        if first_meander_sideways and not self.issideways(
                start_pt_adjusted_up, pts[0], pts[1]):
            pass
        # if start_pt.position is above axes - shift + 2xfillet &  not first_meander_sideways
        elif not first_meander_sideways and self.issideways(
                start_pt_adjusted_down, pts[0], pts[1]):
            pass
        else:
            # else block first mender
            adjustment_vector[:2] = [0, 0]
        # if end_pt.position is below axes + shift - 2xfillet &  last_meander_sideways
        if last_meander_sideways and not self.issideways(
                end_pt_adjusted_up, pts[-2 - term_point], pts[-1 - term_point]):
            pass
        # if end_pt.position is above axes - shift + 2xfillet &  not last_meander_sideways
        elif not last_meander_sideways and self.issideways(
                end_pt_adjusted_down, pts[-2 - term_point],
                pts[-1 - term_point]):
            pass
        else:
            # else block last mender
            adjustment_vector[-2 - term_point:-term_point] = [0, 0]

        not_a_meander = 0
        if term_point:
            # means that pts count is a odd number
            # thus needs to disable shift on the termination point...
            adjustment_vector[-1] = 0
            # ...unless the last point is anchored to the last meander curve
            if start_pt.direction is not None and end_pt.direction is not None:
                if ((mao.dot(start_pt.direction, end_pt.direction) < 0) and
                    (mao.dot(forward, start_pt.direction) <= 0)):
                    # pins are pointing opposite directions and diverging, thus keep consistency
                    adjustment_vector[-1] = adjustment_vector[-2]
                    if adjustment_vector[-1]:
                        # the point in between needs to be shifted, but it will not contribute to length change
                        # therefore the total length distribution (next step) should ignore it.
                        not_a_meander = 1

        # Finally, divide the slack amongst all points...
        sideways_adjustment = sideways * (
            delta_length /
            (np.count_nonzero(adjustment_vector) - not_a_meander))
        pts = pts + sideways_adjustment[
            np.newaxis, :] * adjustment_vector[:, np.newaxis]

        return pts

    @staticmethod
    def get_index_for_side1_meander(num_root_pts: int):
        """Get the indices.

        Args:
            num_root_pts (list): List of points

        Returns:
            tuple: Tuple of indices
        """
        num_2pts, odd = divmod(num_root_pts, 2)

        x = np.array(range(num_2pts), dtype=int) * 4
        z = np.zeros(num_2pts * 2, dtype=int)
        z[::2] = x
        z[1::2] = x + 1
        return z, odd

    def issideways(self, point, seg_point_a, seg_point_b):
        return mao.cross(point - seg_point_a, seg_point_b - seg_point_a) < 0



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Anchored path."""

import numpy as np

from collections import OrderedDict
from qiskit_metal import Dict
from qiskit_metal.qlibrary.core import QRoute, QRoutePoint
from qiskit_metal.toolbox_metal import math_and_overrides as mao
from qiskit_metal.toolbox_metal.exceptions import QiskitMetalDesignError
from collections.abc import Mapping
from shapely.ops import unary_union
from shapely.geometry import CAP_STYLE
import geopandas as gpd


def intersecting(a: np.array, b: np.array, c: np.array, d: np.array) -> bool:
    """Returns whether segment ab intersects or overlaps with segment cd, where
    a, b, c, and d are all coordinates.

    .. meta::
        :description: Anchored Path

    Args:
        a (np.array): Coordinate
        b (np.array): Coordinate
        c (np.array): Coordinate
        d (np.array): Coordinate

    Returns:
        bool: True if intersecting, False otherwise
    """

    x0_start, y0_start = a
    x0_end, y0_end = b
    x1_start, y1_start = c
    x1_end, y1_end = d
    if (x0_start == x0_end) and (x1_start == x1_end):
        # 2 vertical lines intersect only if they completely overlap at some point(s)
        if x0_end == x1_start:
            # Same x-intercept -> potential overlap, so check y coordinate
            # Distinct, non-overlapping segments if and only if min y coord of one is above max y coord of the other
            return not ((min(y0_start, y0_end) > max(y1_start, y1_end)) or
                        (min(y1_start, y1_end) > max(y0_start, y0_end)))
        return False  # Parallel lines with different x-intercepts don't overlap
    elif (x0_start == x0_end) or (x1_start == x1_end):
        # One segment is vertical, the other is not
        # Express non-vertical line in the form of y = mx + b and check y value
        if x1_start == x1_end:
            # Exchange names; the analysis below assumes that line 0 is the vertical one
            x0_start, x0_end, x1_start, x1_end = x1_start, x1_end, x0_start, x0_end
            y0_start, y0_end, y1_start, y1_end = y1_start, y1_end, y0_start, y0_end
        m = (y1_end - y1_start) / (x1_end - x1_start)
        b = (x1_end * y1_start - x1_start * y1_end) / (x1_end - x1_start)
        if min(x1_start, x1_end) <= x0_start <= max(x1_start, x1_end):
            if min(y0_start, y0_end) <= m * x0_start + b <= max(
                    y0_start, y0_end):
                return True
        return False
    else:
        # Neither line is vertical; check slopes and y-intercepts
        b0 = (y0_start * x0_end - y0_end * x0_start) / (
            x0_end - x0_start)  # y-intercept of line 0
        b1 = (y1_start * x1_end - y1_end * x1_start) / (
            x1_end - x1_start)  # y-intercept of line 1
        if (x1_end - x1_start) * (y0_end - y0_start) == (x0_end - x0_start) * (
                y1_end - y1_start):
            # Lines have identical slopes
            if b0 == b1:
                # Same y-intercept -> potential overlap, so check x coordinate
                # Distinct, non-overlapping segments if and only if min x coord of one exceeds max x coord of the other
                return not ((min(x0_start, x0_end) > max(x1_start, x1_end)) or
                            (min(x1_start, x1_end) > max(x0_start, x0_end)))
            return False  # Parallel lines with different y-intercepts don't overlap
        else:
            # Lines not parallel so must intersect somewhere -> examine slopes m0 and m1
            m0 = (y0_end - y0_start) / (x0_end - x0_start)  # slope of line 0
            m1 = (y1_end - y1_start) / (x1_end - x1_start)  # slope of line 1
            x_intersect = (b1 - b0) / (m0 - m1
                                      )  # x coordinate of intersection point
            if min(x0_start, x0_end) <= x_intersect <= max(x0_start, x0_end):
                if min(x1_start, x1_end) <= x_intersect <= max(
                        x1_start, x1_end):
                    return True
            return False


class RouteAnchors(QRoute):
    """Creates and connects a series of anchors through which the Route passes.

    QRoute Default Options:
        * pin_inputs: Dict
            * start_pin: Dict -- Component and pin string pair. Define which pin to start from
                * component: '' -- Name of component to start from, which has a pin
                * pin: '' -- Name of pin used for pin_start
            * end_pin=Dict -- Component and pin string pair. Define which pin to start from
                * component: '' -- Name of component to end on, which has a pin
                * pin: '' -- Name of pin used for pin_end
        * fillet: '0'
        * lead: Dict
            * start_straight: '0mm' -- Lead-in, defined as the straight segment extension from start_pin.  Defaults to 0.1um.
            * end_straight: '0mm' -- Lead-out, defined as the straight segment extension from end_pin.  Defaults to 0.1um.
            * start_jogged_extension: '' -- Lead-in, jogged extension of lead-in. Described as list of tuples
            * end_jogged_extension: '' -- Lead-out, jogged extension of lead-out. Described as list of tuples
        * total_length: '7mm'
        * trace_width: 'cpw_width' -- Defines the width of the line.  Defaults to 'cpw_width'.

    Default Options:
        * anchors: OrderedDict -- Intermediate anchors only; doesn't include endpoints
        * advanced: Dict
            * avoid_collision: 'false' -- true/false, defines if the route needs to avoid collisions.  Defaults to 'false'.
    """

    component_metadata = Dict(short_name='cpw')
    """Component metadata"""

    default_options = Dict(
        anchors=OrderedDict(
        ),  # Intermediate anchors only; doesn't include endpoints
        # Example: {1: np.array([x1, y1]), 2: np.array([x2, y2])}
        # startpin -> startpin + leadin -> anchors -> endpin + leadout -> endpin
        advanced=Dict(avoid_collision='false'))
    """Default options"""

    TOOLTIP = """Creates and connects a series of anchors through which the Route passes."""

    from shapely.ops import unary_union
    from matplotlib import pyplot as plt
    import geopandas as gpd

    from shapely.geometry import CAP_STYLE, JOIN_STYLE

    def unobstructed_close_up(self, segment: list, component_name: str) -> bool:
        """Checks whether the given component's perimeter intersects or
        overlaps a given segment.

        Args:
            segment (list): 2 vertices, in the form [np.array([x0, y0]), np.array([x1, y1])]
            component_name (str): Alphanumeric component name

        Returns:
            bool: True is no obstacles
        """
        # transform path to polygons
        paths_converted = []
        paths = self.design.components[component_name].qgeometry_table('path')
        for _, row in paths.iterrows():
            paths_converted.append(row['geometry'].buffer(
                row['width'] / 2, cap_style=CAP_STYLE.flat))
        # merge all the polygons
        polygons = self.design.components[component_name].qgeometry_list('poly')
        boundary = gpd.GeoSeries(unary_union(polygons + paths_converted))
        boundary_coords = list(boundary.geometry.exterior[0].coords)
        if any(
                intersecting(segment[0], segment[1], boundary_coords[i],
                             boundary_coords[i + 1])
                for i in range(len(boundary_coords) - 1)):
            # At least 1 intersection with the actual component contour; do not proceed!
            return False
        # All clear, no intersections
        return True

    def unobstructed(self, segment: list) -> bool:
        """Check that no component's bounding box in self.design intersects or
        overlaps a given segment.

        Args:
            segment (list): 2 vertices, in the form [np.array([x0, y0]), np.array([x1, y1])]

        Returns:
            bool: True is no obstacles
        """

        # assumes rectangular bounding boxes
        for component in self.design.components:
            if component == self.name:
                continue
            xmin, ymin, xmax, ymax = self.design.components[
                component].qgeometry_bounds()
            # p, q, r, s are corner coordinates of each bounding box
            p, q, r, s = [
                np.array([xmin, ymin]),
                np.array([xmin, ymax]),
                np.array([xmax, ymin]),
                np.array([xmax, ymax])
            ]
            if any(
                    intersecting(segment[0], segment[1], k, l)
                    for k, l in [(p, q), (p, r), (r, s), (q, s)]):
                # At least 1 intersection with the component bounding box. Check the actual contour.
                if not self.unobstructed_close_up(segment, component):
                    # At least 1 intersection with the actual component contour; do not proceed!
                    return False
        # All clear, no intersections
        return True

    def connect_simple(self, start_pt: QRoutePoint,
                       end_pt: QRoutePoint) -> np.ndarray:
        """Try connecting start and end with single or 2-segment/S-shaped CPWs
        if possible.

        Args:
            start_pt (QRoutePoint): QRoutePoint of the start
            end_pt (QRoutePoint): QRoutePoint of the end

        Returns:
            List of vertices of a CPW going from start to end

        Raises:
            QiskitMetalDesignError: If the connect_simple() has failed.
        """
        avoid_collision = self.parse_options().advanced.avoid_collision

        start_direction = start_pt.direction
        start = start_pt.position
        end_direction = end_pt.direction
        end = end_pt.position

        # end_direction originates strictly from endpoint + leadout (NOT intermediate stopping anchors)
        self.assign_direction_to_anchor(start_pt, end_pt)
        stop_direction = end_pt.direction

        if (start[0] == end[0]) or (start[1] == end[1]):
            # Matching x or y coordinates -> check if endpoints can be connected with a single segment
            if mao.dot(start_direction, end - start) >= 0:
                # Start direction and end - start for CPW must not be anti-aligned
                if (end_direction is None) or (mao.dot(end - start,
                                                       end_direction) <= 0):
                    # If leadout + end has been reached, the single segment CPW must not be aligned with its direction
                    return np.empty((0, 2), float)
        else:
            # If the endpoints don't share a common x or y value:
            # designate them as 2 corners of an axis aligned rectangle
            # and check if both start and end directions are aligned with
            # the displacement vectors between start/end and
            # either of the 2 remaining corners ("perfect alignment").
            corner1 = np.array([start[0],
                                end[1]])  # x coordinate matches with start
            corner2 = np.array([end[0],
                                start[1]])  # x coordinate matches with end
            if avoid_collision:
                # Check for collisions at the outset to avoid repeat work
                startc1end = bool(
                    self.unobstructed([start, corner1]) and
                    self.unobstructed([corner1, end]))
                startc2end = bool(
                    self.unobstructed([start, corner2]) and
                    self.unobstructed([corner2, end]))
            else:
                startc1end = startc2end = True
            if (mao.dot(start_direction, corner1 - start) > 0) and startc1end:
                # corner1 is "in front of" the start_pt
                if (end_direction is None) or (mao.dot(end_direction,
                                                       corner1 - end) >= 0):
                    # corner1 is also "in front of" the end_pt
                    return np.expand_dims(corner1, axis=0)
            elif (mao.dot(start_direction, corner2 - start) > 0) and startc2end:
                # corner2 is "in front of" the start_pt
                if (end_direction is None) or (mao.dot(end_direction,
                                                       corner2 - end) >= 0):
                    # corner2 is also "in front of" the end_pt
                    return np.expand_dims(corner2, axis=0)
            # In notation below, corners 3 and 4 correspond to
            # the ends of the segment bisecting the longer rectangle formed by start and end
            # while the segment formed by corners 5 and 6 bisect the shorter rectangle
            if stop_direction[
                    0]:  # "Wide" rectangle -> vertical middle segment is more natural
                corner3 = np.array([(start[0] + end[0]) / 2, start[1]])
                corner4 = np.array([(start[0] + end[0]) / 2, end[1]])
                corner5 = np.array([start[0], (start[1] + end[1]) / 2])
                corner6 = np.array([end[0], (start[1] + end[1]) / 2])
            else:  # "Tall" rectangle -> horizontal middle segment is more natural
                corner3 = np.array([start[0], (start[1] + end[1]) / 2])
                corner4 = np.array([end[0], (start[1] + end[1]) / 2])
                corner5 = np.array([(start[0] + end[0]) / 2, start[1]])
                corner6 = np.array([(start[0] + end[0]) / 2, end[1]])
            if avoid_collision:
                startc3c4end = bool(
                    self.unobstructed([start, corner3]) and
                    self.unobstructed([corner3, corner4]) and
                    self.unobstructed([corner4, end]))
                startc5c6end = bool(
                    self.unobstructed([start, corner5]) and
                    self.unobstructed([corner5, corner6]) and
                    self.unobstructed([corner6, end]))
            else:
                startc3c4end = startc5c6end = True
            if (mao.dot(start_direction, stop_direction) < 0) and (mao.dot(
                    start_direction, corner3 - start) > 0) and startc3c4end:
                if (end_direction is None) or (mao.dot(end_direction,
                                                       corner4 - end) > 0):
                    # Perfectly aligned S-shaped CPW
                    return np.vstack((corner3, corner4))
            # Relax constraints and check if imperfect 2-segment or S-segment works,
            # where "imperfect" means 1 or more dot products of directions
            # between successive segments is 0; otherwise return an empty list
            if (mao.dot(start_direction, corner1 - start) >= 0) and startc1end:
                if (end_direction is None) or (mao.dot(end_direction,
                                                       corner1 - end) >= 0):
                    return np.expand_dims(corner1, axis=0)
            if (mao.dot(start_direction, corner2 - start) >= 0) and startc2end:
                if (end_direction is None) or (mao.dot(end_direction,
                                                       corner2 - end) >= 0):
                    return np.expand_dims(corner2, axis=0)
            if (mao.dot(start_direction, corner3 - start)
                    >= 0) and startc3c4end:
                if (end_direction is None) or (mao.dot(end_direction,
                                                       corner4 - end) >= 0):
                    return np.vstack((corner3, corner4))
            if (mao.dot(start_direction, corner5 - start)
                    >= 0) and startc5c6end:
                if (end_direction is None) or (mao.dot(end_direction,
                                                       corner6 - end) >= 0):
                    return np.vstack((corner5, corner6))
        raise QiskitMetalDesignError(
            "connect_simple() has failed. This might be due to one of two reasons. "
            f"1. Either one of the start point {start} or the end point {end} "
            "provided are inside the bounding box of another QComponent. "
            "Please move the point, or setup a \"lead\" to exit the QComponent area. "
            "2. none of the 4 routing possibilities of this algorithm "
            "(^|_, ^^|, __|, _|^) can complete. Please use Pathfinder instead")

    def free_manhattan_length_anchors(self):
        """Computes the free-flight manhattan distance between start_pt and
        end_pt passing through all of the given anchor points.

        Returns:
            float: Total length connecting all points in order
        """
        anchors = self.parse_options().anchors
        reference = [self.head.get_tip().position]
        reference.extend(list(anchors.values()))
        reference.append(self.tail.get_tip().position)

        length = 0
        for i in range(1, len(reference)):
            length += abs(reference[i][0] -
                          reference[i - 1][0]) + abs(reference[i][1] -
                                                     reference[i - 1][1])
        return length

    def trim_pts(self):
        """Crops the sequence of points to concatenate.

        For example, if a segment between two anchors has no points,
        then the segment is eliminated (only anchor points will do).
        Modified directly the self.intermediate_pts, thus nothing is
        returned.
        """
        if isinstance(self.intermediate_pts, Mapping):
            keys_to_delete = set()
            for key, value in self.intermediate_pts.items():
                if value is None:
                    keys_to_delete.add(key)
                try:
                    # value is a list
                    if not value:
                        keys_to_delete.add(key)
                except ValueError:
                    # value is a numpy
                    if not value.size:
                        keys_to_delete.add(key)
            for key in keys_to_delete:
                del self.intermediate_pts[key]

    def make(self):
        """Generates path from start pin to end pin."""
        p = self.parse_options()
        anchors = p.anchors

        # Set the CPW pins and add the points/directions to the lead-in/out arrays
        self.set_pin("start")
        self.set_pin("end")

        # Align the lead-in/out to the input options set from the user
        start_point = self.set_lead("start")
        end_point = self.set_lead("end")

        self.intermediate_pts = OrderedDict()
        for arc_num, coord in anchors.items():
            arc_pts = self.connect_simple(self.get_tip(), QRoutePoint(coord))
            if arc_pts is None:
                self.intermediate_pts[arc_num] = [coord]
            else:
                self.intermediate_pts[arc_num] = np.concatenate(
                    [arc_pts, [coord]], axis=0)
        arc_pts = self.connect_simple(self.get_tip(), end_point)
        if arc_pts is not None:
            self.intermediate_pts[len(anchors)] = np.array(arc_pts)

        # concatenate all points, transforming the dictionary into a single numpy array
        self.trim_pts()
        self.intermediate_pts = np.concatenate(list(
            self.intermediate_pts.values()),
                                               axis=0)

        # Make points into elements
        self.make_elements(self.get_points())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './elements_ui.ui',
# licensing of './elements_ui.ui' applies.
#
# Created: Thu Jun 30 16:30:20 2022
#      by: pyside6-uic  running on PySide2 5.13.2
#
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore, QtGui, QtWidgets


class Ui_ElementsWindow(object):

    def setupUi(self, ElementsWindow):
        ElementsWindow.setObjectName("ElementsWindow")
        ElementsWindow.resize(841, 623)
        self.centralwidget = QtWidgets.QWidget(ElementsWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setSizeConstraint(
            QtWidgets.QLayout.SetDefaultConstraint)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.btn_refresh = QtWidgets.QPushButton(self.centralwidget)
        self.btn_refresh.setCursor(QtCore.Qt.ClosedHandCursor)
        self.btn_refresh.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/refresh"), QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        self.btn_refresh.setIcon(icon)
        self.btn_refresh.setIconSize(QtCore.QSize(20, 20))
        self.btn_refresh.setAutoDefault(False)
        self.btn_refresh.setDefault(False)
        self.btn_refresh.setFlat(True)
        self.btn_refresh.setObjectName("btn_refresh")
        self.horizontalLayout.addWidget(self.btn_refresh)
        self.label = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum,
                                           QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setLegacyWeight(75)
        font.setBold(True)
        self.label.setFont(font)
        self.label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing |
                                QtCore.Qt.AlignVCenter)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.combo_element_type = QtWidgets.QComboBox(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred,
                                           QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.combo_element_type.sizePolicy().hasHeightForWidth())
        self.combo_element_type.setSizePolicy(sizePolicy)
        self.combo_element_type.setCurrentText("")
        self.combo_element_type.setSizeAdjustPolicy(
            QtWidgets.QComboBox.AdjustToContents)
        self.combo_element_type.setObjectName("combo_element_type")
        self.horizontalLayout.addWidget(self.combo_element_type)
        self.line = QtWidgets.QFrame(self.centralwidget)
        self.line.setFrameShape(QtWidgets.QFrame.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.horizontalLayout.addWidget(self.line)
        self.label_3 = QtWidgets.QLabel(self.centralwidget)
        font = QtGui.QFont()
        font.setLegacyWeight(75)
        font.setBold(True)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout.addWidget(self.label_3)
        self.label_2 = QtWidgets.QLabel(self.centralwidget)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout.addWidget(self.label_2)
        self.lineEdit = QtWidgets.QLineEdit(self.centralwidget)
        self.lineEdit.setObjectName("lineEdit")
        self.horizontalLayout.addWidget(self.lineEdit)
        self.label_4 = QtWidgets.QLabel(self.centralwidget)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout.addWidget(self.label_4)
        self.lineEdit_2 = QtWidgets.QLineEdit(self.centralwidget)
        self.lineEdit_2.setObjectName("lineEdit_2")
        self.horizontalLayout.addWidget(self.lineEdit_2)
        self.line_2 = QtWidgets.QFrame(self.centralwidget)
        self.line_2.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.horizontalLayout.addWidget(self.line_2)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.tableElements = QtWidgets.QTableView(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                           QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.tableElements.sizePolicy().hasHeightForWidth())
        self.tableElements.setSizePolicy(sizePolicy)
        self.tableElements.setProperty("showDropIndicator", False)
        self.tableElements.setDragDropOverwriteMode(False)
        self.tableElements.setAlternatingRowColors(True)
        self.tableElements.setSortingEnabled(False)
        self.tableElements.setObjectName("tableElements")
        self.verticalLayout.addWidget(self.tableElements)
        self.verticalLayout_2.addLayout(self.verticalLayout)
        ElementsWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar()
        self.menubar.setGeometry(QtCore.QRect(0, 0, 841, 24))
        self.menubar.setObjectName("menubar")
        ElementsWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(ElementsWindow)
        self.statusbar.setEnabled(True)
        self.statusbar.setObjectName("statusbar")
        ElementsWindow.setStatusBar(self.statusbar)

        self.retranslateUi(ElementsWindow)
        QtCore.QObject.connect(self.combo_element_type,
                               QtCore.SIGNAL("currentIndexChanged(QString)"),
                               ElementsWindow.combo_element_type)
        QtCore.QObject.connect(self.btn_refresh, QtCore.SIGNAL("clicked()"),
                               ElementsWindow.force_refresh)
        QtCore.QMetaObject.connectSlotsByName(ElementsWindow)

    def retranslateUi(self, ElementsWindow):
        ElementsWindow.setWindowTitle(
            QtWidgets.QApplication.translate("ElementsWindow", "MainWindow",
                                             None, -1))
        self.btn_refresh.setToolTip(
            QtWidgets.QApplication.translate("ElementsWindow",
                                             "Force refresh the table ", None,
                                             -1))
        self.btn_refresh.setStatusTip(
            QtWidgets.QApplication.translate("ElementsWindow",
                                             "Force refresh the table ", None,
                                             -1))
        self.btn_refresh.setWhatsThis(
            QtWidgets.QApplication.translate("ElementsWindow",
                                             "Force refresh the table ", None,
                                             -1))
        self.btn_refresh.setAccessibleDescription(
            QtWidgets.QApplication.translate("ElementsWindow",
                                             "Force refresh the table ", None,
                                             -1))
        self.label.setText(
            QtWidgets.QApplication.translate("ElementsWindow",
                                             "QGeometry type: ", None, -1))
        self.combo_element_type.setToolTip(
            QtWidgets.QApplication.translate(
                "ElementsWindow",
                "<html><head/><body><p>Select the element table you wish to view</p></body></html>",
                None, -1))
        self.label_3.setText(
            QtWidgets.QApplication.translate("ElementsWindow", "  Filter:  ",
                                             None, -1))
        self.label_2.setText(
            QtWidgets.QApplication.translate("ElementsWindow", "Component: ",
                                             None, -1))
        self.label_4.setText(
            QtWidgets.QApplication.translate("ElementsWindow", "  Layer:  ",
                                             None, -1))


from . import main_window_rc_rc



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""QRenderer base class."""

import logging
import inspect
from copy import deepcopy
from typing import TYPE_CHECKING
from typing import List, Tuple, Union, Any, Iterable
from abc import abstractmethod, ABC

from qiskit_metal.designs import is_design
from qiskit_metal.qgeometries import QGeometryTables
from qiskit_metal import designs

from ... import Dict

__all__ = ['QRenderer']

if TYPE_CHECKING:
    # For linting typechecking, import modules that can't be loaded here under normal conditions.
    # For example, I can't import QDesign, because it requires Qrenderer first. We have the
    # chicken and egg issue.
    from qiskit_metal.designs import QDesign


class QRenderer(ABC):
    """Abstract base class for all Renderers of Metal designs and their
	components and qgeometry.

    Handles:
        ::

            designs
                components
                    qgeometry
                        paths
                        polys
                chips
    """

    name = 'base'  # overwrite this!
    """Name"""

    __loaded_renderers__ = set()
    __instantiated_renderers__ = dict()

    # overwrite this to add element extensions:  see ELEMENT_COLUMNS
    # should be dict of dict with keys as element type, which contain (name, dype) pairs
    # e.g. element_extensions = dict(
    #            base=dict(color=str, klayer=int),
    #            path=dict(thickness=float, material=str, perfectE=bool),
    #            poly=dict(thickness=float, material=str), )
    element_extensions = dict()
    """Element extensions dictionary"""

    # TODO: To add: default parameters for the renderer for component element values.
    element_table_data = dict()
    """Element table data."""

    @classmethod
    def load(cls):
        """Load the renderer and register all its extensions. Only performed
        once.

        Once complete, the renderer is added to the class attribute
        '__loaded_renderers__' of QRenderer

        Returns:
            bool: True if success, otherwise throws an error.
        """

        # Check name
        name = cls.name

        if name in QRenderer.__loaded_renderers__:
            pass
            # print(f'Warning: Renderer name={name}, class={cls} already loaded. Doing nothing.')

        cls.populate_element_extensions()

        # Add element extensions
        # see docstring for QRenderer.element_extensions
        QGeometryTables.add_renderer_extension(cls.name, cls.element_extensions)

        # Moved to init for each renderer.
        # Add component extensions

        # to be used in the creation of default params for component qgeometry
        #raise NotImplementedError()

        # Finish and register officially as ready to use.
        QRenderer.__loaded_renderers__.add(name)

        # Reset the table for the next QRenderer.
        for table in cls.element_table_data.keys():
            cls.element_extensions.pop(table, None)

        return True

    @classmethod
    def populate_element_extensions(cls):
        """Populate cls.element_extensions which will be used to create columns
        for tables in QGeometry tables.

        The structure of cls.element_table_data should be same as
        cls.element_extensions.
        """

        for table, a_dict in cls.element_table_data.items():
            cls.element_extensions[table] = dict()
            for col_name, col_value in a_dict.items():
                # type will only tell out about the base class, won't tell you about the inheritance.
                cls.element_extensions[table][col_name] = type(col_value)

    @staticmethod
    def get_renderer(name: str):
        """Returns an already loaded and instantiated renderer.

        Args:
            name (str): rendering name

        Returns:
            QRenderer: Renderer with the given name
        """
        if not name in QRenderer.__loaded_renderers__:
            print(
                'ERROR: The renderer {name} has not yet been loaded. Please use the load function!'
            )

        if not name in QRenderer.__instantiated_renderers__:
            print(
                'ERROR: The renderer {name} has not yet been instantiated. Please instantiate the class!'
            )

        return QRenderer.__instantiated_renderers__[name]

    def __init__(self,
                 design: 'QDesign',
                 initiate=False,
                 render_template: Dict = None,
                 render_options: Dict = None):
        """
        Args:
            design (QDesign): The design
            initiate (bool): True to initiate the renderer.  Defaults to False.
            render_template (Dict, optional): Typically used by GUI for template options for GDS.  Defaults to None.
            render_options (Dict, optional): Used to override all options.  Defaults to None.
        """

        # TODO: check that the renderer has been loaded with load_renderer

        self.status = 'Not Init'

        if design is None:
            print(
                "INFO: A Qiskit Metal design was not provided. Creating an empty design instance"
                " to make the renderer proceed. (Developers: Remove need 4 dummy design, github Issue 631)"
            )
            design = designs.DesignPlanar({}, True)
        assert is_design(
            design), "Error, for the design argument you must provide a\
                                   a child instance of Metal QDesign class."

        self._design = design

        # Options
        self._options = Dict()
        self.update_options(render_options=render_options,
                            render_template=render_template)

        self.initiated = False
        if initiate:
            self.start()

        # Register as an instantiated renderer.
        QRenderer.__instantiated_renderers__[self.name] = self

        self.status = 'Init Completed'

    @property
    def options(self) -> Dict:
        """Options for the QRenderer."""
        return self._options

    @property
    def design(self) -> 'QDesign':
        """Return a reference to the parent design object."""
        return self._design

    @property
    def logger(self) -> logging.Logger:
        """Returns the logger."""
        return self._design.logger

    @classmethod
    def _gather_all_children_default_options(cls) -> Dict:
        """From the base class of QRenderer, traverse the child classes to
        gather the .default_options for each child class.

        Note: If keys are the same for a child and grandchild, the grandchild will
        overwrite the child init method.

        Returns:
            Dict: Options from all children.
        """
        options_from_children = Dict()
        parents = inspect.getmro(cls)

        # QRenderer is not expected to have default_options dict to add to QRenderer class.
        for child in parents[len(parents) - 2::-1]:
            # There is a developer agreement so the defaults for a renderer will be in a dict named default_options.
            if hasattr(child, 'default_options'):
                options_from_children = {
                    **options_from_children,
                    **child.default_options  # pylint: disable=no-member
                }
        return options_from_children

    @classmethod
    def _get_unique_class_name(cls) -> str:
        """Returns unique class name based on the module.

        Returns:
            str: Example: 'qiskit_metal.renders.renderer_gds.gds_renderer.QGDSRenderer'
        """
        return f'{cls.__module__}.{cls.__name__}'

    @classmethod
    def _register_class_with_design(cls, design: 'QDesign', template_key: str,
                                    render_template: Dict):
        """Init function to register a renderer class with the design when
        first instantiated. Registers the renderer's template options.

        Args:
            design (QDesign): The parent design
            template_key (str): Key to use
            render_template (dict): template of render to copy
        """
        # do not overwrite
        if template_key not in design.template_options:
            if not render_template:
                render_template = cls._gather_all_children_default_options()
            design.template_options[template_key] = deepcopy(render_template)

    @classmethod
    def get_template_options(cls,
                             design: 'QDesign',
                             render_template: Dict = None,
                             logger_: logging.Logger = None,
                             template_key: str = None) -> Dict:
        """Creates template options for the Metal QRenderer class required for
        the class to function, based on the design template; i.e., be created,
        made, and rendered. Provides the blank option structure required.

        The options can be extended by plugins, such as renderers.

        Args:
            design (QDesign): A design class.
            render_template (Dict, optional): Template options to overwrite the class ones. Defaults to None.
            logger_ (logging.Logger, optional): A logger for errors. Defaults to None.
            template_key (str, optional): The design.template_options key identifier. If None, then use
                _get_unique_class_name(). Defaults to None.

        Returns:
            Dict: Dictionary of renderer's default options based on design.template_options.
        """

        # get key for templates
        if template_key is None:
            template_key = cls._get_unique_class_name()

        if template_key not in design.template_options:
            # Registers the renderer's template options.
            cls._register_class_with_design(design, template_key,
                                            render_template)

        # Only log warning, if template_key not registered within design.
        if template_key not in design.template_options:
            logger_ = logger_ or design.logger
            if logger_:
                logger_.error(
                    f'ERROR in creating renderer {cls.__name__}!\nThe default '
                    f'options for the renderer class {cls.__name__} are missing'
                )

        # Specific object render template options
        options = deepcopy(Dict(design.template_options[template_key]))

        return options

    def parse_value(self, value: Union[Any, List, Dict, Iterable]) -> Any:
        """Same as design.parse_value. See design for help.

        Returns:
            object: Parsed value of input.
        """
        return self.design.parse_value(value)

    def update_options(self,
                       render_options: Dict = None,
                       render_template: Dict = None):
        """If template options has not been set for this renderer, then gather
        all the default options for children and add to design.  The GUI would
        use this to store the template options.

        Then give the template options to render
        to store in self.options.  Then user can over-ride the render_options.

        Args:
            render_options (Dict, optional): If user wants to over-ride the template
                                             options.  Defaults to None.
            render_template (Dict, optional): All the template options for each child.
                                              Defaults to None.
        """
        self.options.update(
            self.get_template_options(self.design,
                                      render_template=render_template))

        if render_options:
            self.options.update(render_options)

    def add_table_data_to_QDesign(self, class_name: str):
        """During init of renderer, this needs to happen. In particular, each
        renderer needs to update custom columns and values within QDesign.

        Args:
            class_name (str): Name from cls.name for each renderer.
        """
        status = set()
        if not isinstance(QRenderer.name, str):
            self.logger.warning(
                f'In add_table_data_to_QDesign, cls.str={QRenderer.name} is not a str.'
            )
            return

        for table, a_dict in self.element_table_data.items():
            for col_name, col_value in a_dict.items():
                status = self.design.add_default_data_for_qgeometry_tables(
                    table, class_name, col_name, col_value)
                if 5 not in status:
                    self.logger.warning(
                        f'col_value={col_value} not added to QDesign')

    def start(self, force=False):
        """
        Call any initialization (single run) step required to setup the renderer for the first execution,
        such as connecting to some API or COM, or importing the correct material libraries, etc.

        Args:
            force (bool) : If True, need to scrap the existing initialization and re-do
                           If False, will start a new one only if none exists.  Defaults to False.

        Returns:
            bool: is the renderer initialized successfully (thus usable?)
        """

        # TODO: add code here to verify that Ansys is open. if not, then self.initiated=False
        if force or not self.initiated:
            if force and self.initiated:
                # previously initialized renderer, try to shut it down
                self._close_renderer()

        # TODO: move the code line below to inside the `if force or not initiated`,
        #  but only after the TODO before the `if` is completed
        # try to initialize the renderer
        self.initiated = self._initiate_renderer()

        return self.initiated

    def stop(self):
        """
        Any calls that one may want to make after a rendering is complete.
        """
        self.initiated = False
        return self._close_renderer()

    @abstractmethod
    def _initiate_renderer(self):
        """Abstract method. Must be implemented by the subclass.
        Call any initialization (single run) step required to setup the renderer for the first execution,
        such as connecting to some API or COM, or importing the correct material libraries, etc.

        Implementation must return boolean True if succesful. False otherwise.
        """
        return True

    @abstractmethod
    def _close_renderer(self):
        """
        Call any initialization (single run) step required to close the renderer after final execution,
        such as disconnecting from some API or COM, close hanging threads, free memory, etc.

        Implementation must return boolean True if succesful. False otherwise.
        """
        return True

    def get_unique_component_ids(
            self,
            highlight_qcomponents: Union[list,
                                         None] = None) -> Tuple[list, int]:
        """Confirm the list doesn't have names of components repeated. Confirm
        that the name of component exists in QDesign. If QDesign doesn't
        contain any component, or if all components in QDesign are found in
        highlight_qcomponents, return an empty list; otherwise return a list of
        unique components to be sent to the renderer. The second returned item, an
        integer, specifies which of these 3 cases applies.

        Args:
            highlight_qcomponents (Union[list, None], optional): Components to render. Defaults to None.

        Returns:
            Tuple[list, int]: Tuple: Empty or partial list of components in QDesign.
                            int: 0 subset selected
                                 1 every component selected
                                 2 invalid
        """
        highlight_qcomponents = highlight_qcomponents if highlight_qcomponents else []
        unique_qcomponents = set(highlight_qcomponents)
        for qcomp in unique_qcomponents:
            if qcomp not in self.design.name_to_id:
                self.logger.warning(
                    f'The component={qcomp} in highlight_qcomponents not'
                    ' in QDesign.')
                return [], 2  # Invalid
        if len(unique_qcomponents) in (0, len(self.design.components)):
            return [], 1  # Everything selected
        return [self.design.name_to_id[elt] for elt in unique_qcomponents
               ], 0  # Subset selected

    @abstractmethod
    def render_design(self):
        """Abstract method. Must be implemented by the subclass.
        Renders all design chips and components.
        """



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
=================================================
QLibrary (:mod:`qiskit_metal.qlibrary`)
=================================================

.. currentmodule:: qiskit_metal.qlibrary

Module containing all Qiskit Metal QLibrary.

.. _qlibrary:

Core Classes
----------------

.. autosummary::
    :toctree: .

    QComponent
    ParsedDynamicAttributes_Component
    BaseQubit
    QRoute
    QRouteLead
    QRoutePoint


Sample Shapes
-----------------

.. autosummary::
    :toctree:

    CircleCaterpillar
    CircleRaster
    NGon
    NSquareSpiral
    Rectangle
    RectangleHollow


Lumped
----------

.. autosummary::
    :toctree:

    Cap3Interdigital
    CapNInterdigital
    ResonatorCoilRect


Couplers
------------

.. autosummary::
    :toctree:

    CoupledLineTee
    LineTee
    CapNInterdigitalTee
	TunableCoupler01
    TunableCoupler02


Resonator
------------

.. autosummary::
    :toctree:

    ReadoutResFC
    ResonatorLumped


Terminations
----------------

.. autosummary::
    :toctree:

    LaunchpadWirebond
    LaunchpadWirebondCoupled
	LaunchpadWirebondDriven
    OpenToGround
    ShortToGround


Transmission Lines
----------------------

.. autosummary::
    :toctree:

    RouteStraight
    RouteFramed
    RouteMeander
    RouteAnchors
    RouteMixed
    RoutePathfinder


Qubits
----------

.. autosummary::
    :toctree:

    jj_dolan
    jj_manhattan
    TransmonConcentric
    TransmonConcentricType2
    TransmonCross
    TransmonCrossFL
    TransmonInterdigitated
    TransmonPocket
    TransmonPocketCL
    TransmonPocket6
    TransmonPocketTeeth
    TunableCoupler01
    SQUID_LOOP
    StarQubit


Submodules
--------------

.. autosummary::
    :toctree:

    anchored_path

"""

from .core import QComponent
from .core import QRoute
from .core import BaseQubit

from .. import config
if config.is_building_docs():
    from .core import QRouteLead
    from .core import QRoutePoint
    from .core._parsed_dynamic_attrs import ParsedDynamicAttributes_Component
    from .sample_shapes.circle_caterpillar import CircleCaterpillar
    from .sample_shapes.circle_raster import CircleRaster
    from .sample_shapes.n_gon import NGon
    from .sample_shapes.n_square_spiral import NSquareSpiral
    from .sample_shapes.rectangle import Rectangle
    from .sample_shapes.rectangle_hollow import RectangleHollow
    from .couplers.coupled_line_tee import CoupledLineTee
    from .couplers.line_tee import LineTee
    from .couplers.cap_n_interdigital_tee import CapNInterdigitalTee
    from .couplers.tunable_coupler_01 import TunableCoupler01
    from .couplers.tunable_coupler_02 import TunableCoupler02
    from .lumped.cap_n_interdigital import CapNInterdigital
    from .lumped.cap_3_interdigital import Cap3Interdigital
    from .lumped.resonator_coil_rect import ResonatorCoilRect
    from .terminations.launchpad_wb import LaunchpadWirebond
    from .terminations.launchpad_wb_coupled import LaunchpadWirebondCoupled
    from .terminations.launchpad_wb_driven import LaunchpadWirebondDriven
    from .terminations.open_to_ground import OpenToGround
    from .terminations.short_to_ground import ShortToGround
    from .tlines.straight_path import RouteStraight
    from .tlines.framed_path import RouteFramed
    from .tlines.meandered import RouteMeander
    from .tlines.anchored_path import RouteAnchors
    from .tlines.mixed_path import RouteMixed
    from .tlines.pathfinder import RoutePathfinder
    from .qubits.JJ_Dolan import jj_dolan
    from .qubits.JJ_Manhattan import jj_manhattan
    from .qubits.transmon_concentric import TransmonConcentric
    from .qubits.transmon_concentric_type_2 import TransmonConcentricType2
    from .qubits.transmon_cross import TransmonCross
    from .qubits.transmon_cross_fl import TransmonCrossFL
    from .qubits.Transmon_Interdigitated import TransmonInterdigitated
    from .qubits.transmon_pocket import TransmonPocket
    from .qubits.transmon_pocket_cl import TransmonPocketCL
    from .qubits.transmon_pocket_6 import TransmonPocket6
    from .qubits.transmon_pocket_teeth import TransmonPocketTeeth
    from .qubits.SQUID_loop import SQUID_LOOP
    from .qubits.star_qubit import StarQubit
    from .resonators.readoutres_fc import ReadoutResFC
    from .resonators.resonator_lumped import ResonatorLumped

    from .tlines import anchored_path



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Tree model for Param Entry Window
"""

import builtins
import json
import queue
from collections import OrderedDict
from typing import Union, TYPE_CHECKING, Any

import numpy as np
from PySide6 import QtCore
from PySide6.QtCore import QAbstractItemModel, QModelIndex, QTimer, Qt
from PySide6.QtGui import QFont
from PySide6.QtWidgets import (QComboBox, QTreeView, QWidget)
from addict import Dict

if TYPE_CHECKING:
    from qiskit_metal.designs.design_base import QDesign


def get_nested_dict_item(dic: dict, key_list: list):
    """
    Get a nested dictionary item.
    If key_list is empty, return dic itself.

    Args:
        dic (dict): dictionary of items
        key_list (list): list of keys

    Returns:
        dict: nested dictionary

    .. code-block:: python
        :linenos:

        myDict = Dict(aa=Dict(x1={'dda':34},y1='Y',z='10um'),
            bb=Dict(x2=5,y2='YYYsdg',z='100um'))
        key_list = ['aa', 'x1', 'dda']
        [get_dict_item](myDict, key_list)
        returns 34

    """
    if key_list:
        for k in key_list:
            if k not in dic:
                raise Exception(f"CANAOT FIND ERROR{k} not in {dic}!!! :( ")
            dic = dic[k]

    return dic


class Node:  # pylint: disable=too-few-public-methods
    """
    This class was made for type hints (instead of having to union BranchNode and LeafNode)"
    """
    KEY, NODE = range(2)


class BranchNode(Node):
    """
    A BranchNode object has a nonzero number of child nodes.
    These child nodes can be either BranchNodes or LeafNodes.

    It is uniquely defined by its name, parent node, and list of children.
    The list of children consists of tuple pairs of the form (nodename_i, node_i),
    where the former is the name of the child node and the latter is the child node itself.
    KEY (=0) and NODE (=1) identify their respective positions within each tuple pair.

    """

    def __init__(self, name: str, parent=None, cur_type=dict):
        """
        Args:
            name (str): Name of this branch
            parent ([type]): The parent (Default: None)
            data (dict): Node data (Default: None)
        """
        super().__init__()
        self.name = name
        self.parent = parent
        self.children = []
        self.type = cur_type.__name__

    def update_name(self, new_name):
        """ Updates name """
        self.name = new_name
        self.parent.update_child(self)

    def get_type_combobox(self, parent: QWidget):
        """Gets ComboBox"""

        return self.BranchTypeComboBox(parent)

    class BranchTypeComboBox(QComboBox):
        """
        Class for holding the different branch types user can choose from
        """
        orderedDict = OrderedDict.__name__
        rdict = dict.__name__
        addict = Dict.__name__

        branch_type_names = {orderedDict, rdict, addict}
        branch_types = {OrderedDict, dict, Dict}
        type_dictionary = {rdict: dict, orderedDict: OrderedDict, addict: Dict}
        """
        Contains all possible types for values in a DEB
        """

        def __init__(self, parent=None):
            """ Inits Combobox"""
            super().__init__(parent)
            self.setAutoFillBackground(True)  # must be set for adding to tree
            for btn in self.branch_type_names:
                self.addItem(btn)

        def getType(self):  # pylint: disable=invalid-name disable=inconsistent-return-statements
            """ Return Type """
            if self.currentText() in self.type_dictionary:
                return self.type_dictionary[self.currentText()]

        def getTypeName(self):  # pylint: disable=invalid-name
            """ Return Name"""
            return self.currentText()

    def get_empty_dictionary(self):
        """ Returns correct empty dict"""
        if self.type == self.BranchTypeComboBox.orderedDict:
            return OrderedDict()
        return dict()

    def __len__(self):
        """
        Gets the number of children

        Returns:
            int: the number of children this node has
        """
        return len(self.children)

    def __str__(self):
        """String of object"""
        mystr = f"BranchNode: {self.name} with children: "
        for chi in self.children:
            mystr += f"{chi[0]}, "

        return mystr

    def childAtRow(self, row: int):  # pylint: disable=invalid-name disable=inconsistent-return-statements
        """Gets the child at the given row

        Args:
            row (int): the row

        Returns:
            Node: The node at the row
        """
        if 0 <= row < len(self.children):
            return self.children[row][self.NODE]

    def rowOfChild(self, child):  # pylint: disable=invalid-name
        """Gets the row of the given child

        Args:
            child (Node): the child

        Returns:
            int: Row of the given child.  -1 is returned if the child is not found.
        """
        for i, (_, childnode) in enumerate(self.children):
            if child == childnode:
                return i
        return -1

    def childWithKey(self, key: str):  # pylint: disable=invalid-name
        """Gets the child with the given key

        Args:
            key (str): the key

        Returns:
            Node: The child with the same name as the given key.
            None is returned if the child is not found
        """
        for childname, childnode in self.children:
            if key == childname:
                return childnode
        return None

    def insertChild(self, child):  # pylint: disable=invalid-name
        """
        Insert the given child

        Args:
            child (Node): the child
        """
        child.parent = self
        self.children.append((child.name, child))

    def update_child(self, child_node):
        """ Updates child"""
        row = self.rowOfChild(child_node)
        self.children[row] = (child_node.name, child_node)

    def hasLeaves(self):  # pylint: disable=invalid-name
        """Do I have leaves?

        Returns:
            bool: True is I have leaves, False otherwise
        """
        if not self.children:
            return False
        return isinstance(self.children[self.KEY][self.NODE], LeafNode)


class NpEncoder(json.JSONEncoder):
    """Helper Class for data transfer"""

    # https://docs.python.org/3/library/json.html
    def default(self, obj):  # pylint: disable=arguments-differ
        """Returns default"""
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super().default(obj)


class LeafNode(Node):
    """
    A LeafNode object has no children but consists of a key-value pair, denoted by
    name and value, respectively.

    It is uniquely identified by its root-to-leaf path, which is a list of keys
    whose positions denote their nesting depth (shallow to deep).
    """

    def __init__(self, name: str, value: Any, parent=None):
        """
        Args:
            name (str): Label for the leaf node
            value: (Any): Value for the leaf node
            parent (Node): The parent (Default: None)
        """
        super().__init__()
        self.parent = parent
        self.name = name
        self.type = type(value).__name__
        self.value = self._get_display_value(value)

    def get_type_combobox(self, parent: QWidget):
        """Return combobox"""
        return self.LeafTypeComboBox(parent)

    class LeafTypeComboBox(QComboBox):
        """
        Contains all possible types for values in a DEB
        """
        npArr = "ndarray"
        pylist = "list"
        customFromStringItems = {
            npArr: lambda a: np.array(json.loads(a)),
            pylist: lambda a: json.loads(a),  # pylint: disable=unnecessary-lambda
        }
        customToStringItems = {
            npArr: lambda a: json.dumps(a, cls=NpEncoder),
            pylist: lambda a: json.dumps(a, cls=NpEncoder),
        }

        def __init__(self, parent=None):
            """ Inits Leaf Combobox"""
            super().__init__(parent)
            self.setAutoFillBackground(True)  # must be set for adding to tree
            self.addItem("str")
            self.addItem("float")
            self.addItem("int")
            self.addItem("bool")

            self.addItem(self.npArr)
            self.addItem(self.pylist)
            self.np_enc = NpEncoder

        def getType(self):  # pylint: disable=invalid-name
            """Return type"""
            if self.currentText() in self.customFromStringItems:

                return self.customFromStringItems[self.currentText()]

            return getattr(builtins, self.currentText())

        def getTypeName(self):  # pylint: disable=invalid-name
            """Return name"""
            return self.currentText()

    def _get_display_value(self, value):
        """Return display value """

        if self.type in self.LeafTypeComboBox.customToStringItems:
            return self.LeafTypeComboBox.customToStringItems[self.type](value)
        return str(value)

    def get_real_value(self):
        """Return real value"""
        if self.type in self.LeafTypeComboBox.customFromStringItems:
            val = self.LeafTypeComboBox.customFromStringItems[self.type](
                self.value)
            return val

        cur_type = getattr(builtins, self.type)
        if cur_type == bool:  # pylint: disable=simplifiable-if-statement
            if self.value.lower() in "true":  # pylint: disable=simplifiable-if-statement
                value = True
            else:
                value = False
        else:
            value = cur_type(self.value)
        return value

    def update_name(self, new_name):
        """Update name"""
        self.name = new_name
        self.parent.update_child(self)

    def __str__(self):
        """ To str"""
        mystr = f"LeafNode: {self.name} with value: {self.value}"
        return mystr


class TreeModelParamEntry(QAbstractItemModel):
    """
    Tree model for a general hierarchical dataset.

    This class extends the `QAbstractItemModel` class.
    It is part of the model-view-controller (MVC) architecture; see
    https://doc.qt.io/qt-5/qabstractitemmodel.html

    Access using ``gui.component_window.model``.
    """

    __refreshtime = 500  # 0.5 second refresh time
    NAME = 0
    VALUE = 1
    PARSED = 2
    TYPE = 3

    def __init__(self,
                 parent: QWidget,
                 view: QTreeView,
                 data_dict: OrderedDict = None,
                 design: 'QDesign' = None):
        """
        Editable table with drop-down rows for a generic qcomponent menu.
        Organized as a tree model where child nodes are more specific properties
        of a given parent node.

        Args:
            parent (QWidget): The parent widget
            view (QTreeView):  Model's related QTreeView
            data_dict (OrderedDict): Data to be stored in the model's tree
            design (QDesign): Current design using the model
        """
        super().__init__(parent=parent)
        self._row_count = -1  # pylint: disable=invalid-name
        self.root = BranchNode('')
        self.view = view
        self._design = design
        self.headers = ['Name', 'Value']
        self._start_timer()
        self.headers = ['Name', 'Value', 'Parsed Value',
                        "Type"]  # 3 columns instead of 2
        if data_dict is None:
            data_dict = {}
        self.init_load(data_dict)

    def add_new_leaf_node(self,
                          cur_index: QModelIndex,
                          key: str,
                          value: Any = None):
        """
        Add new leaf node to model's backing tree structure
        Args:
            cur_index (QModelIndex): Index of parent of new leaf
            key (str): Key value in leaf node's key-value pair
            value (Any): Value in lead node's key-value pair

        Returns:

        """
        parent_node = self.nodeFromIndex(cur_index)

        if isinstance(parent_node, LeafNode):
            parent_node = parent_node.parent

        parent_node.insertChild(LeafNode(key, value, parent=parent_node))

        cop = self.get_path_of_expanded_items()

        self.reload()
        self.expand_items_in_paths(cop)

    def add_new_branch_node(self,
                            cur_index: QModelIndex,
                            key: str,
                            fake_key: str,
                            fake_value: Any = None):
        """
        Adds new branch node to model's backing tree
        Args:
            cur_index (QModelIndex): Index of parent of new branch
            key (str): Key value, aka, name of new branch
            fake_key (str): Placeholder leaf's key since QTreeView cannot display childless BranchNodes
            fake_value (Any): Placeholder leaf's value since QTreeView cant display childless BranchNodes

        """
        parent_node = self.nodeFromIndex(cur_index)
        if isinstance(parent_node, LeafNode):
            parent_node = parent_node.parent

        new_node = BranchNode(key, parent_node)
        new_node.insertChild(LeafNode(fake_key, fake_value, new_node))
        parent_node.insertChild(new_node)
        cop = self.get_path_of_expanded_items()

        self.reload()
        self.expand_items_in_paths(cop)

    def delete_node(self, cur_index: QModelIndex):
        """
        Removes node from backing tree
        Args:
            cur_index (QModelIndex): Index of node to be removed from backing tree

        Returns:

        """
        cur_node = self.nodeFromIndex(cur_index)

        if cur_node == self.root:

            raise Exception("Don't delete the everything please")

        # remove from parent
        if (cur_node.name, cur_node) in cur_node.parent.children:

            cur_node.parent.children.remove((cur_node.name, cur_node))

            # cleaning up dangling branches
            parent = old_parent = cur_node.parent
            if len(parent.children) < 1:
                while len(parent.children) < 2:
                    old_parent = parent
                    parent = old_parent.parent
                    old_parent.children = []

                if len(old_parent.children) < 1:
                    parent.children.remove((old_parent.name, old_parent))

        else:
            raise Exception(
                "Internal Model Exception: unable to find parent of child node to be deleted."
            )

        cop = self.get_path_of_expanded_items()

        self.dataChanged.emit(cur_index, cur_index)
        self.reload()
        self.expand_items_in_paths(cop)

    def _start_timer(self):
        """
        Start and continuously refresh timer in background to keep
        the total number of rows up to date.
        """
        self.timer = QTimer(self)
        self.timer.start(self.__refreshtime)
        self.timer.timeout.connect(self.auto_refresh)

    def auto_refresh(self):
        """
        Check to see if the total number of rows has been changed. If so,
        completely rebuild the model and tree.
        """
        # TODO: Check if new nodes have been added; if so, rebuild model.
        new_row_count = self.rowCount(self.createIndex(0, 0))  # pylint: disable=invalid-name
        if self._row_count != new_row_count:
            # Wrap the reset logic in beginResetModel and endResetModel
            self.beginResetModel()
            try:

                # When a model is reset it should be considered that all
                # information previously retrieved from it is invalid.
                # This includes but is not limited to the rowCount() and
                # columnCount(), flags(), data retrieved through data(), and roleNames().
                # This will loose the current selection.
                # self.modelReset.emit()

                self._row_count = new_row_count
            finally:
                self.endResetModel()

    def getPaths(self, curdict: OrderedDict, curpath: list):  # pylint: disable=invalid-name
        """Recursively finds and saves all root-to-leaf paths in model"""

        for k, v in curdict.items():
            if type(v) in BranchNode.BranchTypeComboBox.branch_types:
                self.getPaths(v, curpath + [k])
            else:
                self.paths.append(curpath + [k, v])

    def reload(self):
        """ Sends out a signal forcing QTreeView to completely refresh"""
        self.beginResetModel()
        self.endResetModel()

    def init_load(self, data_dict: dict):
        """Builds a tree from a dictionary (data_dict)"""

        if len(data_dict) < 1:
            return
        self.beginResetModel()

        # Clear existing tree paths if any
        self.root.children.clear()

        # Construct the paths -> sets self.paths
        self.paths = []
        self.getPaths(data_dict, [])
        for path in self.paths:
            root = self.root
            branch = None
            # Combine final name and value for leaf node,
            # so stop at 2nd to last element of each path
            for key in path[:-2]:
                # Look for childnode with the name 'key'. If it's not found, create a new branch.
                branch = root.childWithKey(key)
                if not branch:
                    dict_type = type(
                        get_nested_dict_item(data_dict,
                                             path[:path.index(key) + 1]))

                    branch = BranchNode(key, parent=root, cur_type=dict_type)
                    root.insertChild(branch)
                root = branch
            # If a LeafNode resides in the outermost dictionary, the above for loop is bypassed.
            # [Note: This happens when the root-to-leaf path length is just 2.]
            # In this case, add the LeafNode right below the master root.
            # Otherwise, add the LeafNode below the final branch.
            root.insertChild(LeafNode(path[-2], path[-1], parent=root))

        # Emit a signal since the model's internal state
        # (e.g. persistent model indexes) has been invalidated.
        self.endResetModel()

    def rowCount(self, parent: QModelIndex):
        """Get the number of rows

        Args:
            parent (QModelIndex): The parent

        Returns:
            int: The number of rows
        """
        # If there is no component, then show placeholder text
        if len(self.root.children) < 1:
            return 0

        # Count number of child nodes belonging to the parent.
        node = self.nodeFromIndex(parent)
        if (node is None) or isinstance(node, LeafNode):
            return 0

        return len(node)

    def columnCount(self, parent: QModelIndex = None):  # pylint: disable=unused-argument
        """Get the number of columns

        Args:
            parent (QModelIndex): The parent

        Returns:
            int: The number of columns
        """
        return len(self.headers)

    def data(self, index: QModelIndex, role: Qt.ItemDataRole = Qt.DisplayRole):  # pylint: disable=too-many-return-statements,too-many-branches
        """Gets the node data

        Args:
            index (QModelIndex): index to get data for
            role (Qt.ItemDataRole): the role (Default: Qt.DisplayRole)

        Returns:
            object: fetched data
        """
        if not index.isValid():
            return None

        # The data in a form suitable for editing in an editor. (QString)
        if role == Qt.EditRole:
            return self.data(index, Qt.DisplayRole)

        # Bold the first
        if (role == Qt.FontRole) and (index.column() == self.NAME):
            font = QFont()
            font.setBold(True)
            return font

        if role == Qt.TextAlignmentRole:
            return int(Qt.AlignTop | Qt.AlignLeft)

        if role == Qt.DisplayRole:
            node = self.nodeFromIndex(index)
            if node:
                # the first column is either a leaf key or a branch
                # the second column is always a leaf value or for a branch is ''.
                if isinstance(node, BranchNode):
                    # Handle a branch (which is a nested subdictionary, which can be expanded)
                    if index.column() == self.NAME:
                        return node.name
                    if index.column() == self.TYPE:
                        return node.type
                    return ''
                # We have a leaf
                if index.column() == self.NAME:
                    return str(node.name)  # key
                if index.column() == self.VALUE:
                    return str(node.value)  # value
                if index.column() == self.PARSED:
                    return str(self._design.parse_value(node.value))
                if index.column() == self.TYPE:
                    return node.type

                return None

        return None

    def setData(
            self,  # pylint: disable=too-many-return-statements
            index: QModelIndex,
            value: Any,
            role: Qt.ItemDataRole = Qt.EditRole) -> bool:
        """Set the LeafNode value and corresponding data entry to value.
        Returns true if successful; otherwise returns false.
        The dataChanged() signal should be emitted if the data was successfully set.

        Args:
            index (QModelIndex): The index
            value (Any): The value
            role (Qt.ItemDataRole): The role of the data (Default: Qt.EditRole)

        Returns:
            bool: True if successful, False otherwise
        """
        try:

            if not index.isValid():
                return False

            if role == QtCore.Qt.EditRole:

                if index.column(
                ) == self.VALUE:  # only want to edit col1 if it's a leafnode

                    node = self.nodeFromIndex(index)

                    if isinstance(node, LeafNode):

                        value = str(value)  # new value
                        old_value = node.value  # option value

                        if old_value == value:
                            return False

                        # Set the value of an option when the new value is different
                        node.value = value  # # pylint: disable=attribute-defined-outside-init
                        return True  # why doesn't this require a self.init_load()?

                elif index.column(
                ) == self.NAME:  # either editing BranchNode name or LeafNode name
                    node = self.nodeFromIndex(index)
                    if node.parent is None:
                        raise Exception("Trying to edit node without a parent. "
                                        "The root node should be uneditable.")

                    old_value = node.name
                    if old_value == value:
                        return False

                    node.update_name(value)

                    cop = self.get_path_of_expanded_items()

                    self.dataChanged.emit(index, index)
                    self.expand_items_in_paths(cop)

                    # update node
                    # at each parent:
                    # data.emit changed from 0,0 to len(children), lenCol
                    return True
                elif index.column() == self.TYPE:
                    # this is a type
                    node = self.nodeFromIndex(
                        self.index(index.row(), 0, index.parent()))
                    node.type = value

                    cop = self.get_path_of_expanded_items()

                    self.dataChanged.emit(index, index)
                    self.expand_items_in_paths(cop)
                    return True

            return False
        except Exception as e:  # pylint: disable=broad-except
            self._design.logger.error(f"Unable to parse tree information: {e}")
            return False

    # persistentIndexList
    def get_path_of_expanded_items(self):
        """ Collect which nodes are currently expanded in
        order to re-expand them after refreshing the model"""
        expanded_nodes = []
        pil = self.persistentIndexList()
        for mil in pil:
            is_expanded = self.view.isExpanded(mil)
            if is_expanded:
                minode = self.nodeFromIndex(mil)
                expanded_nodes.append(minode)
        return expanded_nodes

    def expand_items_in_paths(self, expanded_nodes):
        """Expand all nodes that were previously saved in expanded_nodes"""
        INDEX = 0  # pylint: disable=invalid-name
        NODE = 1  # pylint: disable=invalid-name
        cur_index = QModelIndex()
        node = self.nodeFromIndex(cur_index)
        discovered_queue = queue.Queue()
        discovered_queue.put((cur_index, node))

        while discovered_queue.qsize() > 0:
            to_check = discovered_queue.get()
            if to_check[NODE] in expanded_nodes:

                self.view.setExpanded(to_check[INDEX], True)

            # find all children and add them to queue
            parent_index = to_check[INDEX]
            new_index = self.index(0, 0, parent_index)  # be valid
            row = 0
            while (new_index is not None and new_index.isValid()):

                new_index = self.index(row, 0, parent_index)
                if new_index is not None and new_index.isValid():

                    new_node = self.nodeFromIndex(new_index)
                    if isinstance(new_node, BranchNode):
                        discovered_queue.put((new_index, new_node))

                    row = row + 1

    def headerData(self, section: int, orientation: Qt.Orientation,
                   role: Qt.ItemDataRole):
        """ Set the headers to be displayed.

        Args:
            section (int): section number
            orientation (Qt.Orientation): the orientation
            role (Qt.ItemDataRole): the role

        Returns:
            object: header data
        """
        if orientation == Qt.Horizontal:
            if role == Qt.DisplayRole:
                if 0 <= section < len(self.headers):
                    return self.headers[section]

            elif role == Qt.FontRole:
                font = QFont()
                font.setBold(True)
                return font

        return None

    def index(self, row: int, column: int, parent: QModelIndex):
        """What is my index?

        Args:
            row (int): the row
            column (int): the column
            parent (QModelIndex): the parent

        Returns:
            int: internal index
        """
        assert self.root
        branch = self.nodeFromIndex(parent)
        assert branch is not None
        if isinstance(branch, BranchNode):
            # The third argument is the internal index.
            node_data = branch.childAtRow(row)
            if node_data is not None:
                return self.createIndex(row, column, node_data)

        return self.createIndex(
            -1, -1, QModelIndex())  # return invalid index which is Branch("")

    def parent(self, child: QModelIndex):
        """Gets the parent index of the given node

        Args:
            child (node): the child

        Returns:
            int: the index
        """
        node = self.nodeFromIndex(child)

        if node is None:
            return QModelIndex()

        parent = node.parent
        if parent is None:
            return QModelIndex()

        grandparent = parent.parent
        if grandparent is None:
            return QModelIndex()
        row = grandparent.rowOfChild(parent)
        assert row != -1
        return self.createIndex(row, 0, parent)

    def nodeFromIndex(self, index: QModelIndex) -> Union[BranchNode, LeafNode]:  # pylint: disable=invalid-name
        """Utility method we define to get the node from the index.

        Args:
            index (QModelIndex): The model index

        Returns:
            Union[BranchNode, LeafNode]: Returns the node, which can either be
            a branch (for a dict) or a leaf (for an option key value pairs)
        """
        if index.isValid():
            # The internal pointer will return the leaf or branch node under the given parent.
            return index.internalPointer()
        return self.root

    def flags(self, index: QModelIndex):
        """
        Determine how user may interact with each cell in the table.

        Returns:
            list: flags
        """
        flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled
        if index.column() == self.NAME or index.column() == self.TYPE:
            return flags | Qt.ItemIsEditable

        if index.column() == self.VALUE:
            node = self.nodeFromIndex(index)
            if isinstance(node, LeafNode):
                return flags | Qt.ItemIsEditable

        return flags

    def node_str(self, node: Node):  # pylint: disable=no-self-use
        """Node to string"""
        if Node is None:
            return "None"
        return str(node)
